import os
from functools import cached_property
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from pydantic import ConfigDict
from tenacity import retry, stop_after_attempt, wait_random_exponential
from tenacity.retry import retry_if_not_exception_type

if TYPE_CHECKING:
    from redisvl.extensions.cache.embeddings.embeddings import EmbeddingsCache

from redisvl.utils.vectorize.base import BaseVectorizer


class VertexAIVectorizer(BaseVectorizer):
    """The VertexAIVectorizer uses Google's VertexAI embedding model
    API to create embeddings.

    This vectorizer is tailored for use in
    environments where integration with Google Cloud Platform (GCP) services is
    a key requirement.

    Utilizing this vectorizer requires an active GCP project and location
    (region), along with appropriate application credentials. These can be
    provided through the `api_config` dictionary or set the GOOGLE_APPLICATION_CREDENTIALS
    env var. Additionally, the vertexai python client must be
    installed with `pip install google-cloud-aiplatform>=1.26`.

    You can optionally enable caching to improve performance when generating
    embeddings for repeated inputs.

    .. code-block:: python

        # Basic usage
        vectorizer = VertexAIVectorizer(
            model="textembedding-gecko",
            api_config={
                "project_id": "your_gcp_project_id", # OR set GCP_PROJECT_ID
                "location": "your_gcp_location",     # OR set GCP_LOCATION
            })
        embedding = vectorizer.embed("Hello, world!")

        # With caching enabled
        from redisvl.extensions.cache.embeddings import EmbeddingsCache
        cache = EmbeddingsCache(name="vertexai_embeddings_cache")

        vectorizer = VertexAIVectorizer(
            model="textembedding-gecko",
            api_config={
                "project_id": "your_gcp_project_id",
                "location": "your_gcp_location",
            },
            cache=cache
        )

        # First call will compute and cache the embedding
        embedding1 = vectorizer.embed("Hello, world!")

        # Second call will retrieve from cache
        embedding2 = vectorizer.embed("Hello, world!")

        # Batch embedding of multiple texts
        embeddings = vectorizer.embed_many(
            ["Hello, world!", "Goodbye, world!"],
            batch_size=2
        )

        # Multimodal usage
        from vertexai.vision_models import Image, Video

        vectorizer = VertexAIVectorizer(
            model="multimodalembedding@001",
            api_config={
                "project_id": "your_gcp_project_id", # OR set GCP_PROJECT_ID
                "location": "your_gcp_location",     # OR set GCP_LOCATION
            }
        )
        text_embedding = vectorizer.embed("Hello, world!")
        image_embedding = vectorizer.embed(Image.load_from_file("path/to/your/image.jpg"))
        video_embedding = vectorizer.embed(Video.load_from_file("path/to/your/video.mp4"))

    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(
        self,
        model: str = "textembedding-gecko",
        api_config: Optional[Dict] = None,
        dtype: str = "float32",
        cache: Optional["EmbeddingsCache"] = None,
        **kwargs,
    ):
        """Initialize the VertexAI vectorizer.

        Args:
            model (str): Model to use for embedding. Defaults to
                'textembedding-gecko'.
            api_config (Optional[Dict], optional): Dictionary containing the
                API config details. Defaults to None.
            dtype (str): the default datatype to use when embedding text as byte arrays.
                Used when setting `as_buffer=True` in calls to embed() and embed_many().
                Defaults to 'float32'.
            cache (Optional[EmbeddingsCache]): Optional EmbeddingsCache instance to cache embeddings for
                better performance with repeated texts. Defaults to None.

        Raises:
            ImportError: If the google-cloud-aiplatform library is not installed.
            ValueError: If the API key is not provided.
            ValueError: If an invalid dtype is provided.
        """
        super().__init__(model=model, dtype=dtype, cache=cache)
        # Initialize client and set up the model
        self._setup(api_config, **kwargs)

    @property
    def is_multimodal(self) -> bool:
        """Whether a multimodal model has been configured."""
        return "multimodal" in self.model

    @cached_property
    def _client(self):
        """Get the appropriate client based on the model type."""
        if self.is_multimodal:
            from vertexai.vision_models import MultiModalEmbeddingModel

            return MultiModalEmbeddingModel.from_pretrained(self.model)

        from vertexai.language_models import TextEmbeddingModel

        return TextEmbeddingModel.from_pretrained(self.model)

    def embed_image(self, image_path: str, **kwargs) -> Union[List[float], bytes]:
        """Embed an image (from its path on disk) using a VertexAI multimodal model."""
        if not self.is_multimodal:
            raise ValueError("Cannot embed image with a non-multimodal model.")

        from vertexai.vision_models import Image

        return self.embed(Image.load_from_file(image_path), **kwargs)

    def embed_video(self, video_path: str, **kwargs) -> Union[List[float], bytes]:
        """Embed a video (from its path on disk) using a VertexAI multimodal model."""
        if not self.is_multimodal:
            raise ValueError("Cannot embed video with a non-multimodal model.")

        from vertexai.vision_models import Video

        return self.embed(Video.load_from_file(video_path), **kwargs)

    def _setup(self, api_config: Optional[Dict], **kwargs):
        """Set up the VertexAI client and determine the embedding dimensions."""
        # Initialize client
        self._initialize_client(api_config, **kwargs)
        # Set model dimensions after initialization
        self.dims = self._set_model_dims()

    def _initialize_client(self, api_config: Optional[Dict], **kwargs):
        """
        Setup the VertexAI client using the provided config options or
        environment variables.

        Args:
            api_config: Dictionary with GCP configuration options
            **kwargs: Additional arguments for initialization

        Raises:
            ImportError: If the google-cloud-aiplatform library is not installed
            ValueError: If required parameters are not provided
        """
        # Fetch the project_id and location from api_config or environment variables
        project_id = (
            api_config.get("project_id") if api_config else os.getenv("GCP_PROJECT_ID")
        )
        location = (
            api_config.get("location") if api_config else os.getenv("GCP_LOCATION")
        )

        if not project_id:
            raise ValueError(
                "Missing project_id. "
                "Provide the id in the api_config with key 'project_id' "
                "or set the GCP_PROJECT_ID environment variable."
            )

        if not location:
            raise ValueError(
                "Missing location. "
                "Provide the location (region) in the api_config with key 'location' "
                "or set the GCP_LOCATION environment variable."
            )

        # Check for credentials
        credentials = api_config.get("credentials") if api_config else None

        try:
            import vertexai

            vertexai.init(
                project=project_id, location=location, credentials=credentials
            )

        except ImportError:
            raise ImportError(
                "VertexAI vectorizer requires the google-cloud-aiplatform library. "
                "Please install with `pip install google-cloud-aiplatform>=1.26`"
            )

    def _set_model_dims(self) -> int:
        """
        Determine the dimensionality of the embedding model by making a test call.

        Returns:
            int: Dimensionality of the embedding model

        Raises:
            ValueError: If embedding dimensions cannot be determined
        """
        try:
            # Call the protected _embed method to avoid caching this test embedding
            embedding = self._embed("dimension check")
            return len(embedding)
        except (KeyError, IndexError) as ke:
            raise ValueError(f"Unexpected response from the VertexAI API: {str(ke)}")
        except Exception as e:  # pylint: disable=broad-except
            # fall back (TODO get more specific)
            raise ValueError(f"Error setting embedding model dimensions: {str(e)}")

    @retry(
        wait=wait_random_exponential(min=1, max=60),
        stop=stop_after_attempt(6),
        retry=retry_if_not_exception_type((TypeError, ValueError)),
    )
    def _embed(self, content: Any, **kwargs) -> List[float]:
        """
        Generate a vector embedding for a single input using the VertexAI API.

        Args:
            content: Input to embed
            **kwargs: Additional parameters to pass to the VertexAI API

        Returns:
            List[float]: Vector embedding as a list of floats

        Raises:
            ValueError: If embedding fails
        """
        try:
            if self.is_multimodal:
                from vertexai.vision_models import Image, Video

                if isinstance(content, str):
                    result = self._client.get_embeddings(
                        contextual_text=content,
                        **kwargs,
                    )
                    if result.text_embedding is None:
                        raise ValueError("No text embedding returned from VertexAI.")
                    return result.text_embedding
                elif isinstance(content, Image):
                    result = self._client.get_embeddings(
                        image=content,
                        **kwargs,
                    )
                    if result.image_embedding is None:
                        raise ValueError("No image embedding returned from VertexAI.")
                    return result.image_embedding
                elif isinstance(content, Video):
                    result = self._client.get_embeddings(
                        video=content,
                        **kwargs,
                    )
                    if result.video_embeddings is None:
                        raise ValueError("No video embedding returned from VertexAI.")
                    return result.video_embeddings[0].embedding
                else:
                    raise TypeError(
                        "Invalid input type for multimodal embedding. "
                        "Must be str, Image, or Video."
                    )

            else:
                return self._client.get_embeddings([content], **kwargs)[0].values

        except Exception as e:
            raise ValueError(f"Embedding input failed: {e}")

    @retry(
        wait=wait_random_exponential(min=1, max=60),
        stop=stop_after_attempt(6),
        retry=retry_if_not_exception_type((TypeError, ValueError)),
    )
    def _embed_many(
        self, contents: List[str], batch_size: int = 10, **kwargs
    ) -> List[List[float]]:
        """
        Generate vector embeddings for a batch of texts using the VertexAI API.

        Args:
            contents: List of texts to embed
            batch_size: Number of texts to process in each API call
            **kwargs: Additional parameters to pass to the VertexAI API

        Returns:
            List[List[float]]: List of vector embeddings as lists of floats

        Raises:
            ValueError: If embedding fails
        """
        if self.is_multimodal:
            raise NotImplementedError(
                "Batch embedding is not supported for multimodal models with VertexAI."
            )
        if not isinstance(contents, list):
            raise TypeError("Must pass in a list of str values to embed.")
        if contents and not isinstance(contents[0], str):
            raise TypeError("Must pass in a list of str values to embed.")

        try:
            embeddings: List = []
            for batch in self.batchify(contents, batch_size):
                response = self._client.get_embeddings(batch, **kwargs)
                embeddings.extend([r.values for r in response])
            return embeddings
        except Exception as e:
            raise ValueError(f"Embedding texts failed: {e}")

    def _serialize_for_cache(self, content: Any) -> Union[bytes, str]:
        """Convert content to a cacheable format."""
        from vertexai.vision_models import Image, Video

        if isinstance(content, Image):
            return content._image_bytes
        elif isinstance(content, Video):
            return content._video_bytes
        return super()._serialize_for_cache(content)

    @property
    def type(self) -> str:
        return "vertexai"



"""Base LLM cache interface for RedisVL.

This module defines the abstract base interface for LLM caches, which store
prompt-response pairs with semantic retrieval capabilities.
"""

from typing import Any, Dict, List, Optional

from redisvl.extensions.cache.base import BaseCache
from redisvl.query.filter import FilterExpression


class BaseLLMCache(BaseCache):
    """Base abstract LLM cache interface.

    This class defines the core functionality for caching LLM responses
    with semantic similarity search capabilities.
    """

    def __init__(self, name: str, ttl: Optional[int] = None, **kwargs):
        """Initialize an LLM cache.

        Args:
            name (str): The name of the cache.
            ttl (Optional[int]): The time-to-live for cached responses. Defaults to None.
            **kwargs: Additional arguments passed to the parent class.
        """
        super().__init__(name=name, ttl=ttl, **kwargs)

    def delete(self) -> None:
        """Delete the cache and its index entirely."""
        raise NotImplementedError

    async def adelete(self) -> None:
        """Async delete the cache and its index entirely."""
        raise NotImplementedError

    def check(
        self,
        prompt: Optional[str] = None,
        vector: Optional[List[float]] = None,
        num_results: int = 1,
        return_fields: Optional[List[str]] = None,
        filter_expression: Optional[FilterExpression] = None,
        distance_threshold: Optional[float] = None,
    ) -> List[Dict[str, Any]]:
        """Check the cache for semantically similar prompts.

        Args:
            prompt (Optional[str]): The text prompt to search for in the cache.
            vector (Optional[List[float]]): Vector representation to search for.
            num_results (int): Number of results to return. Defaults to 1.
            return_fields (Optional[List[str]]): Fields to return in results.
            filter_expression (Optional[FilterExpression]): Optional filter to apply.
            distance_threshold (Optional[float]): Override for semantic distance threshold.

        Returns:
            List[Dict[str, Any]]: List of matching cache entries.
        """
        raise NotImplementedError

    async def acheck(
        self,
        prompt: Optional[str] = None,
        vector: Optional[List[float]] = None,
        num_results: int = 1,
        return_fields: Optional[List[str]] = None,
        filter_expression: Optional[FilterExpression] = None,
        distance_threshold: Optional[float] = None,
    ) -> List[Dict[str, Any]]:
        """Async check the cache for semantically similar prompts."""
        raise NotImplementedError

    def store(
        self,
        prompt: str,
        response: str,
        vector: Optional[List[float]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        filters: Optional[Dict[str, Any]] = None,
        ttl: Optional[int] = None,
    ) -> str:
        """Store a prompt-response pair in the cache.

        Args:
            prompt (str): The user prompt to cache.
            response (str): The LLM response to cache.
            vector (Optional[List[float]]): Optional embedding vector.
            metadata (Optional[Dict[str, Any]]): Optional metadata.
            filters (Optional[Dict[str, Any]]): Optional filters for retrieval.
            ttl (Optional[int]): Optional TTL override.

        Returns:
            str: The Redis key for the cached entry.
        """
        raise NotImplementedError

    async def astore(
        self,
        prompt: str,
        response: str,
        vector: Optional[List[float]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        filters: Optional[Dict[str, Any]] = None,
        ttl: Optional[int] = None,
    ) -> str:
        """Async store a prompt-response pair in the cache."""
        raise NotImplementedError

    def update(self, key: str, **kwargs) -> None:
        """Update specific fields within an existing cache entry.

        Args:
            key (str): The key of the document to update.
            **kwargs: Field-value pairs to update.
        """
        raise NotImplementedError

    async def aupdate(self, key: str, **kwargs) -> None:
        """Async update specific fields within an existing cache entry."""
        raise NotImplementedError



import warnings

import pytest
from redis.exceptions import ConnectionError

from redisvl.extensions.constants import ID_FIELD_NAME
from redisvl.extensions.message_history import MessageHistory, SemanticMessageHistory
from tests.conftest import SKIP_HF, skip_if_no_redisearch

requires_hf = pytest.mark.skipif(
    SKIP_HF, reason="sentence-transformers not supported on Python 3.14+"
)


@pytest.fixture
def app_name():
    return "test_app"


@pytest.fixture
def standard_history(app_name, client):
    history = MessageHistory(app_name, redis_client=client)
    yield history
    history.clear()


@pytest.fixture
def semantic_history(app_name, client, hf_vectorizer):
    skip_if_no_redisearch(client)
    history = SemanticMessageHistory(
        app_name, redis_client=client, overwrite=True, vectorizer=hf_vectorizer
    )
    yield history
    history.clear()
    history.delete()


@pytest.fixture(autouse=True)
def disable_deprecation_warnings():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        yield


# test standard message history
def test_specify_redis_client(client):
    history = MessageHistory(name="test_app", redis_client=client)
    assert isinstance(history._index.client, type(client))


def test_specify_redis_url(client, redis_url):
    history = MessageHistory(
        name="test_app",
        session_tag="abc",
        redis_url=redis_url,
    )
    assert isinstance(history._index.client, type(client))


def test_standard_bad_connection_info():
    with pytest.raises(ConnectionError):
        MessageHistory(
            name="test_app",
            session_tag="abc",
            redis_url="redis://localhost:6389",  # bad url
        )


def test_standard_store(standard_history):
    context = standard_history.get_recent()
    assert len(context) == 0

    standard_history.store(prompt="first prompt", response="first response")
    standard_history.store(prompt="second prompt", response="second response")
    standard_history.store(prompt="third prompt", response="third response")
    standard_history.store(prompt="fourth prompt", response="fourth response")
    standard_history.store(prompt="fifth prompt", response="fifth response")

    # test that order is maintained
    full_context = standard_history.get_recent(top_k=10)
    assert full_context == [
        {"role": "user", "content": "first prompt"},
        {"role": "llm", "content": "first response"},
        {"role": "user", "content": "second prompt"},
        {"role": "llm", "content": "second response"},
        {"role": "user", "content": "third prompt"},
        {"role": "llm", "content": "third response"},
        {"role": "user", "content": "fourth prompt"},
        {"role": "llm", "content": "fourth response"},
        {"role": "user", "content": "fifth prompt"},
        {"role": "llm", "content": "fifth response"},
    ]


def test_standard_add_and_get(standard_history):
    context = standard_history.get_recent()
    assert len(context) == 0

    standard_history.add_message({"role": "user", "content": "first prompt"})
    standard_history.add_message({"role": "llm", "content": "first response"})
    standard_history.add_message({"role": "user", "content": "second prompt"})
    standard_history.add_message({"role": "llm", "content": "second response"})
    standard_history.add_message(
        {
            "role": "tool",
            "content": "tool result 1",
            "tool_call_id": "tool call one",
            "metadata": {"tool call params": "abc 123"},
        }
    )
    standard_history.add_message(
        {
            "role": "tool",
            "content": "tool result 2",
            "tool_call_id": "tool call two",
            "metadata": {"tool call params": "abc 456"},
        }
    )
    standard_history.add_message({"role": "user", "content": "third prompt"})
    standard_history.add_message({"role": "llm", "content": "third response"})

    # test default context history size
    default_context = standard_history.get_recent()
    assert len(default_context) == 5  #  default is 5

    # test specified context history size
    partial_context = standard_history.get_recent(top_k=3)
    assert len(partial_context) == 3
    assert partial_context == [
        {
            "role": "tool",
            "content": "tool result 2",
            "tool_call_id": "tool call two",
            "metadata": {"tool call params": "abc 456"},
        },
        {"role": "user", "content": "third prompt"},
        {"role": "llm", "content": "third response"},
    ]

    # test that order is maintained
    full_context = standard_history.get_recent(top_k=10)
    assert full_context == [
        {"role": "user", "content": "first prompt"},
        {"role": "llm", "content": "first response"},
        {"role": "user", "content": "second prompt"},
        {"role": "llm", "content": "second response"},
        {
            "role": "tool",
            "content": "tool result 1",
            "tool_call_id": "tool call one",
            "metadata": {"tool call params": "abc 123"},
        },
        {
            "role": "tool",
            "content": "tool result 2",
            "tool_call_id": "tool call two",
            "metadata": {"tool call params": "abc 456"},
        },
        {"role": "user", "content": "third prompt"},
        {"role": "llm", "content": "third response"},
    ]

    # test that a ValueError is raised when top_k is invalid
    with pytest.raises(ValueError):
        bad_context = standard_history.get_recent(top_k=-2)

    with pytest.raises(ValueError):
        bad_context = standard_history.get_recent(top_k=-2.0)

    with pytest.raises(ValueError):
        bad_context = standard_history.get_recent(top_k=1.3)

    with pytest.raises(ValueError):
        bad_context = standard_history.get_recent(top_k="3")


def test_standard_add_messages(standard_history):
    context = standard_history.get_recent()
    assert len(context) == 0

    standard_history.add_messages(
        [
            {"role": "user", "content": "first prompt"},
            {
                "role": "llm",
                "content": "first response",
                "metadata": {"llm provider": "openai"},
            },
            {"role": "user", "content": "second prompt"},
            {"role": "llm", "content": "second response"},
            {
                "role": "tool",
                "content": "tool result 1",
                "tool_call_id": "tool call one",
            },
            {
                "role": "tool",
                "content": "tool result 2",
                "tool_call_id": "tool call two",
            },
            {"role": "user", "content": "fourth prompt"},
            {"role": "llm", "content": "fourth response"},
        ]
    )

    full_context = standard_history.get_recent(top_k=10)
    assert len(full_context) == 8
    assert full_context == [
        {"role": "user", "content": "first prompt"},
        {
            "role": "llm",
            "content": "first response",
            "metadata": {"llm provider": "openai"},
        },
        {"role": "user", "content": "second prompt"},
        {"role": "llm", "content": "second response"},
        {"role": "tool", "content": "tool result 1", "tool_call_id": "tool call one"},
        {"role": "tool", "content": "tool result 2", "tool_call_id": "tool call two"},
        {"role": "user", "content": "fourth prompt"},
        {"role": "llm", "content": "fourth response"},
    ]


def test_standard_messages_property(standard_history):
    standard_history.add_messages(
        [
            {"role": "user", "content": "first prompt"},
            {"role": "llm", "content": "first response"},
            {"role": "user", "content": "second prompt"},
            {
                "role": "llm",
                "content": "second response",
                "metadata": {"params": "abc"},
            },
            {"role": "user", "content": "third prompt", "metadata": 42},
        ]
    )

    assert standard_history.messages == [
        {"role": "user", "content": "first prompt"},
        {"role": "llm", "content": "first response"},
        {"role": "user", "content": "second prompt"},
        {"role": "llm", "content": "second response", "metadata": {"params": "abc"}},
        {"role": "user", "content": "third prompt", "metadata": 42},
    ]


def test_standard_scope(standard_history):
    # store entries under default session tag
    standard_history.store("some prompt", "some response")

    # test that changing session tag does indeed change access scope
    new_session = "def"
    standard_history.store(
        "new user prompt", "new user response", session_tag=new_session
    )
    context = standard_history.get_recent(session_tag=new_session)
    assert context == [
        {"role": "user", "content": "new user prompt"},
        {"role": "llm", "content": "new user response"},
    ]

    # test that default session data is still accessible
    context = standard_history.get_recent()
    assert context == [
        {"role": "user", "content": "some prompt"},
        {"role": "llm", "content": "some response"},
    ]

    bad_session = "xyz"
    no_context = standard_history.get_recent(session_tag=bad_session)
    assert no_context == []


def test_standard_get_text(standard_history):
    standard_history.store("first prompt", "first response")
    text = standard_history.get_recent(as_text=True)
    assert text == ["first prompt", "first response"]

    standard_history.add_message({"role": "system", "content": "system level prompt"})
    text = standard_history.get_recent(as_text=True)
    assert text == ["first prompt", "first response", "system level prompt"]


def test_standard_get_raw(standard_history):
    standard_history.store("first prompt", "first response")
    standard_history.store("second prompt", "second response")
    raw = standard_history.get_recent(raw=True)
    assert len(raw) == 4
    assert raw[0]["role"] == "user"
    assert raw[0]["content"] == "first prompt"
    assert raw[1]["role"] == "llm"
    assert raw[1]["content"] == "first response"


def test_standard_drop(standard_history):
    standard_history.store("first prompt", "first response")
    standard_history.store("second prompt", "second response")
    standard_history.store("third prompt", "third response")
    standard_history.store("fourth prompt", "fourth response")

    # test drop() with no arguments removes the last element
    standard_history.drop()
    context = standard_history.get_recent(top_k=3)
    assert context == [
        {"role": "user", "content": "third prompt"},
        {"role": "llm", "content": "third response"},
        {"role": "user", "content": "fourth prompt"},
    ]

    # test drop(id) removes the specified element
    context = standard_history.get_recent(top_k=10, raw=True)
    middle_id = context[3][ID_FIELD_NAME]
    standard_history.drop(middle_id)
    context = standard_history.get_recent(top_k=6)
    assert context == [
        {"role": "user", "content": "first prompt"},
        {"role": "llm", "content": "first response"},
        {"role": "user", "content": "second prompt"},
        {"role": "user", "content": "third prompt"},
        {"role": "llm", "content": "third response"},
        {"role": "user", "content": "fourth prompt"},
    ]


def test_standard_clear(standard_history):
    standard_history.store("some prompt", "some response")
    standard_history.clear()
    empty_context = standard_history.get_recent(top_k=10)
    assert empty_context == []


# test semantic message history
@requires_hf
def test_semantic_specify_client(client, hf_vectorizer):
    skip_if_no_redisearch(client)
    history = SemanticMessageHistory(
        name="test_app",
        session_tag="abc",
        redis_client=client,
        overwrite=True,
        vectorizer=hf_vectorizer,
    )
    assert isinstance(history._index.client, type(client))


@requires_hf
def test_semantic_bad_connection_info(hf_vectorizer):
    with pytest.raises(ConnectionError):
        SemanticMessageHistory(
            name="test_app",
            session_tag="abc",
            redis_url="redis://localhost:6389",
            vectorizer=hf_vectorizer,
        )


@requires_hf
def test_semantic_scope(semantic_history):
    # store entries under default session tag
    semantic_history.store("some prompt", "some response")

    # test that changing session tag does indeed change access scope
    new_session = "def"
    semantic_history.store(
        "new user prompt", "new user response", session_tag=new_session
    )
    context = semantic_history.get_recent(session_tag=new_session)
    assert context == [
        {"role": "user", "content": "new user prompt"},
        {"role": "llm", "content": "new user response"},
    ]

    # test that previous session data is still accessible
    context = semantic_history.get_recent()
    assert context == [
        {"role": "user", "content": "some prompt"},
        {"role": "llm", "content": "some response"},
    ]

    bad_session = "xyz"
    no_context = semantic_history.get_recent(session_tag=bad_session)
    assert no_context == []


@requires_hf
def test_semantic_store_and_get_recent(semantic_history):
    context = semantic_history.get_recent()
    assert len(context) == 0

    semantic_history.store(prompt="first prompt", response="first response")
    semantic_history.store(prompt="second prompt", response="second response")
    semantic_history.store(prompt="third prompt", response="third response")
    semantic_history.store(prompt="fourth prompt", response="fourth response")
    semantic_history.add_message(
        {"role": "tool", "content": "tool result", "tool_call_id": "tool id"}
    )
    semantic_history.add_message(
        {
            "role": "tool",
            "content": "tool result",
            "tool_call_id": "tool id",
            "metadata": "return value from tool",
        }
    )  # test default context history size
    default_context = semantic_history.get_recent()
    assert len(default_context) == 5  # 5 is default

    # test specified context history size
    partial_context = semantic_history.get_recent(top_k=4)
    assert len(partial_context) == 4

    # test larger context history returns full history
    too_large_context = semantic_history.get_recent(top_k=100)
    assert len(too_large_context) == 10

    # test that order is maintained
    full_context = semantic_history.get_recent(top_k=10)
    assert full_context == [
        {"role": "user", "content": "first prompt"},
        {"role": "llm", "content": "first response"},
        {"role": "user", "content": "second prompt"},
        {"role": "llm", "content": "second response"},
        {"role": "user", "content": "third prompt"},
        {"role": "llm", "content": "third response"},
        {"role": "user", "content": "fourth prompt"},
        {"role": "llm", "content": "fourth response"},
        {"role": "tool", "content": "tool result", "tool_call_id": "tool id"},
        {
            "role": "tool",
            "content": "tool result",
            "tool_call_id": "tool id",
            "metadata": "return value from tool",
        },
    ]

    # test that more recent entries are returned
    context = semantic_history.get_recent(top_k=4)
    assert context == [
        {"role": "user", "content": "fourth prompt"},
        {"role": "llm", "content": "fourth response"},
        {"role": "tool", "content": "tool result", "tool_call_id": "tool id"},
        {
            "role": "tool",
            "content": "tool result",
            "tool_call_id": "tool id",
            "metadata": "return value from tool",
        },
    ]

    # test no entries are returned and no error is raised if top_k == 0
    context = semantic_history.get_recent(top_k=0)
    assert context == []

    # test that a ValueError is raised when top_k is invalid
    with pytest.raises(ValueError):
        bad_context = semantic_history.get_recent(top_k=0.5)

    with pytest.raises(ValueError):
        bad_context = semantic_history.get_recent(top_k=-1)

    with pytest.raises(ValueError):
        bad_context = semantic_history.get_recent(top_k=-2.0)

    with pytest.raises(ValueError):
        bad_context = semantic_history.get_recent(top_k=1.3)

    with pytest.raises(ValueError):
        bad_context = semantic_history.get_recent(top_k="3")


@requires_hf
def test_semantic_messages_property(semantic_history):
    semantic_history.add_messages(
        [
            {"role": "user", "content": "first prompt"},
            {"role": "llm", "content": "first response"},
            {
                "role": "tool",
                "content": "tool result 1",
                "tool_call_id": "tool call one",
                "metadata": 42,
            },
            {
                "role": "tool",
                "content": "tool result 2",
                "tool_call_id": "tool call two",
                "metadata": [1, 2, 3],
            },
            {"role": "user", "content": "second prompt"},
            {"role": "llm", "content": "second response"},
            {"role": "user", "content": "third prompt"},
        ]
    )

    assert semantic_history.messages == [
        {"role": "user", "content": "first prompt"},
        {"role": "llm", "content": "first response"},
        {
            "role": "tool",
            "content": "tool result 1",
            "tool_call_id": "tool call one",
            "metadata": 42,
        },
        {
            "role": "tool",
            "content": "tool result 2",
            "tool_call_id": "tool call two",
            "metadata": [1, 2, 3],
        },
        {"role": "user", "content": "second prompt"},
        {"role": "llm", "content": "second response"},
        {"role": "user", "content": "third prompt"},
    ]


@requires_hf
def test_semantic_add_and_get_relevant(semantic_history):
    semantic_history.add_message(
        {"role": "system", "content": "discussing common fruits and vegetables"}
    )
    semantic_history.store(
        prompt="list of common fruits",
        response="apples, oranges, bananas, strawberries",
    )
    semantic_history.store(
        prompt="list of common vegetables",
        response="carrots, broccoli, onions, spinach",
    )
    semantic_history.store(
        prompt="winter sports in the olympics",
        response="downhill skiing, ice skating, luge",
    )
    semantic_history.add_message(
        {
            "role": "tool",
            "content": "skiing, skating, luge",
            "tool_call_id": "winter_sports()",
        }
    )

    # test default distance metric
    default_context = semantic_history.get_relevant(
        "set of common fruits like apples and bananas"
    )
    assert len(default_context) == 2
    assert default_context[0] == {"role": "user", "content": "list of common fruits"}
    assert default_context[1] == {
        "role": "llm",
        "content": "apples, oranges, bananas, strawberries",
    }

    # test increasing distance metric broadens results
    semantic_history.set_distance_threshold(0.5)
    default_context = semantic_history.get_relevant("list of fruits and vegetables")
    assert len(default_context) == 5  # 2 pairs of prompt:response, and system
    assert default_context == semantic_history.get_relevant(
        "list of fruits and vegetables", distance_threshold=0.5
    )

    # test tool calls can also be returned
    context = semantic_history.get_relevant("winter sports like skiing")
    assert context == [
        {
            "role": "user",
            "content": "winter sports in the olympics",
        },
        {
            "role": "tool",
            "content": "skiing, skating, luge",
            "tool_call_id": "winter_sports()",
        },
        {
            "role": "llm",
            "content": "downhill skiing, ice skating, luge",
        },
    ]

    # test that a ValueError is raised when top_k is invalid
    with pytest.raises(ValueError):
        bad_context = semantic_history.get_relevant("test prompt", top_k=-1)

    with pytest.raises(ValueError):
        bad_context = semantic_history.get_relevant("test prompt", top_k=-2.0)

    with pytest.raises(ValueError):
        bad_context = semantic_history.get_relevant("test prompt", top_k=1.3)

    with pytest.raises(ValueError):
        bad_context = semantic_history.get_relevant("test prompt", top_k="3")


@requires_hf
def test_semantic_get_raw(semantic_history):
    semantic_history.store("first prompt", "first response")
    semantic_history.store("second prompt", "second response")
    raw = semantic_history.get_recent(raw=True)
    assert len(raw) == 4
    assert raw[0]["role"] == "user"
    assert raw[0]["content"] == "first prompt"
    assert raw[1]["role"] == "llm"
    assert raw[1]["content"] == "first response"


@requires_hf
def test_semantic_drop(semantic_history):
    semantic_history.store("first prompt", "first response")
    semantic_history.store("second prompt", "second response")
    semantic_history.store("third prompt", "third response")
    semantic_history.store("fourth prompt", "fourth response")

    # test drop() with no arguments removes the last element
    semantic_history.drop()
    context = semantic_history.get_recent(top_k=3)
    assert context == [
        {"role": "user", "content": "third prompt"},
        {"role": "llm", "content": "third response"},
        {"role": "user", "content": "fourth prompt"},
    ]

    # test drop(id) removes the specified element
    context = semantic_history.get_recent(top_k=5, raw=True)
    middle_id = context[2][ID_FIELD_NAME]
    semantic_history.drop(middle_id)
    context = semantic_history.get_recent(top_k=4)
    assert context == [
        {"role": "user", "content": "second prompt"},
        {"role": "llm", "content": "second response"},
        {"role": "llm", "content": "third response"},
        {"role": "user", "content": "fourth prompt"},
    ]


def test_different_vector_dtypes(client, redis_url):
    skip_if_no_redisearch(client)
    try:
        bfloat_sess = SemanticMessageHistory(
            name="bfloat_history", dtype="bfloat16", redis_url=redis_url
        )
        bfloat_sess.add_message({"role": "user", "content": "bfloat message"})

        float16_sess = SemanticMessageHistory(
            name="float16_history", dtype="float16", redis_url=redis_url
        )
        float16_sess.add_message({"role": "user", "content": "float16 message"})

        float32_sess = SemanticMessageHistory(
            name="float32_history", dtype="float32", redis_url=redis_url
        )
        float32_sess.add_message({"role": "user", "content": "float32 message"})

        float64_sess = SemanticMessageHistory(
            name="float64_history", dtype="float64", redis_url=redis_url
        )
        float64_sess.add_message({"role": "user", "content": "float64 message"})

        for sess in [bfloat_sess, float16_sess, float32_sess, float64_sess]:
            sess.set_distance_threshold(0.7)
            assert len(sess.get_relevant("float message")) == 1
            sess.delete()  # Clean up
    except:
        pytest.skip("Required Redis modules not available or version too low")


def test_bad_dtype_connecting_to_exiting_history(client, redis_url):
    skip_if_no_redisearch(client)
    # Skip this test for Redis 6.2.x as FT.INFO doesn't return dims properly
    redis_version = client.info()["redis_version"]
    if redis_version.startswith("6.2"):
        pytest.skip(
            "Redis 6.2.x FT.INFO doesn't properly return vector dims for reconnection"
        )

    def create_history():
        return SemanticMessageHistory(
            name="float64 history", dtype="float64", redis_url=redis_url
        )

    def create_same_type():
        return SemanticMessageHistory(
            name="float64 history", dtype="float64", redis_url=redis_url
        )

    history = create_history()
    same_type = create_same_type()
    # under the hood uses from_existing

    with pytest.raises(ValueError):
        bad_type = SemanticMessageHistory(
            name="float64 history", dtype="float16", redis_url=redis_url
        )


@requires_hf
def test_vectorizer_dtype_mismatch(client, redis_url, hf_vectorizer_float16):
    skip_if_no_redisearch(client)
    with pytest.raises(ValueError):
        SemanticMessageHistory(
            name="test_dtype_mismatch",
            dtype="float32",
            vectorizer=hf_vectorizer_float16,
            redis_url=redis_url,
            overwrite=True,
        )


def test_invalid_vectorizer(client, redis_url):
    skip_if_no_redisearch(client)
    with pytest.raises(TypeError):
        SemanticMessageHistory(
            name="test_invalid_vectorizer",
            vectorizer="invalid_vectorizer",  # type: ignore
            redis_url=redis_url,
            overwrite=True,
        )


def test_passes_through_dtype_to_default_vectorizer(client, redis_url):
    skip_if_no_redisearch(client)
    # The default is float32, so we should see float64 if we pass it in.
    cache = SemanticMessageHistory(
        name="test_pass_through_dtype",
        dtype="float64",
        redis_url=redis_url,
        overwrite=True,
    )
    assert cache._vectorizer.dtype == "float64"


def test_deprecated_dtype_argument(client, redis_url):
    skip_if_no_redisearch(client)
    with pytest.warns(DeprecationWarning):
        SemanticMessageHistory(
            name="float64 history", dtype="float64", redis_url=redis_url, overwrite=True
        )



from redisvl.schema.fields import (
    BaseField,
    CompressionType,
    FieldTypes,
    FlatVectorField,
    GeoField,
    HNSWVectorField,
    NumericField,
    SVSVectorField,
    TagField,
    TextField,
    VectorDataType,
    VectorDistanceMetric,
    VectorIndexAlgorithm,
)
from redisvl.schema.schema import IndexInfo, IndexSchema, StorageType

# Expose validation functionality
from redisvl.schema.validation import validate_object

__all__ = [
    "IndexSchema",
    "IndexInfo",
    "StorageType",
    "FieldTypes",
    "VectorDistanceMetric",
    "VectorDataType",
    "VectorIndexAlgorithm",
    "CompressionType",
    "BaseField",
    "TextField",
    "TagField",
    "NumericField",
    "GeoField",
    "FlatVectorField",
    "HNSWVectorField",
    "SVSVectorField",
    "validate_object",
]



import json

import pytest
from pydantic import ValidationError

from redisvl.extensions.cache.llm import CacheEntry, CacheHit
from redisvl.redis.utils import array_to_buffer, hashify


def test_valid_cache_entry_creation():
    entry = CacheEntry(
        prompt="What is AI?",
        response="AI is artificial intelligence.",
        prompt_vector=[0.1, 0.2, 0.3],
    )
    assert entry.entry_id == hashify("What is AI?")
    assert entry.prompt == "What is AI?"
    assert entry.response == "AI is artificial intelligence."
    assert entry.prompt_vector == [0.1, 0.2, 0.3]


def test_cache_entry_with_given_entry_id():
    entry = CacheEntry(
        entry_id="custom_id",
        prompt="What is AI?",
        response="AI is artificial intelligence.",
        prompt_vector=[0.1, 0.2, 0.3],
    )
    assert entry.entry_id == "custom_id"


def test_cache_entry_with_invalid_metadata():
    with pytest.raises(ValidationError):
        CacheEntry(
            prompt="What is AI?",
            response="AI is artificial intelligence.",
            prompt_vector=[0.1, 0.2, 0.3],
            metadata="invalid_metadata",
        )


def test_cache_entry_to_dict():
    entry = CacheEntry(
        prompt="What is AI?",
        response="AI is artificial intelligence.",
        prompt_vector=[0.1, 0.2, 0.3],
        metadata={"author": "John"},
        filters={"category": "technology"},
    )
    result = entry.to_dict(dtype="float32")
    assert result["entry_id"] == hashify("What is AI?", {"category": "technology"})
    assert result["metadata"] == json.dumps({"author": "John"})
    assert result["prompt_vector"] == array_to_buffer([0.1, 0.2, 0.3], "float32")
    assert result["category"] == "technology"
    assert "filters" not in result


def test_valid_cache_hit_creation():
    hit = CacheHit(
        entry_id="entry_1",
        prompt="What is AI?",
        response="AI is artificial intelligence.",
        vector_distance=0.1,
        inserted_at=1625819123.123,
        updated_at=1625819123.123,
    )
    assert hit.entry_id == "entry_1"
    assert hit.prompt == "What is AI?"
    assert hit.response == "AI is artificial intelligence."
    assert hit.vector_distance == 0.1
    assert hit.inserted_at == hit.updated_at == 1625819123.123


def test_cache_hit_with_serialized_metadata():
    hit = CacheHit(
        entry_id="entry_1",
        prompt="What is AI?",
        response="AI is artificial intelligence.",
        vector_distance=0.1,
        inserted_at=1625819123.123,
        updated_at=1625819123.123,
        metadata=json.dumps({"author": "John"}),
    )
    assert hit.metadata == {"author": "John"}


def test_cache_hit_to_dict():
    hit = CacheHit(
        entry_id="entry_1",
        prompt="What is AI?",
        response="AI is artificial intelligence.",
        vector_distance=0.1,
        inserted_at=1625819123.123,
        updated_at=1625819123.123,
        filters={"category": "technology"},
    )
    result = hit.to_dict()
    assert result["entry_id"] == "entry_1"
    assert result["prompt"] == "What is AI?"
    assert result["response"] == "AI is artificial intelligence."
    assert result["vector_distance"] == 0.1
    assert result["category"] == "technology"
    assert "filters" not in result


def test_cache_entry_with_empty_optional_fields():
    entry = CacheEntry(
        prompt="What is AI?",
        response="AI is artificial intelligence.",
        prompt_vector=[0.1, 0.2, 0.3],
    )
    result = entry.to_dict(dtype="float32")
    assert "metadata" not in result
    assert "filters" not in result


def test_cache_hit_with_empty_optional_fields():
    hit = CacheHit(
        entry_id="entry_1",
        prompt="What is AI?",
        response="AI is artificial intelligence.",
        vector_distance=0.1,
        inserted_at=1625819123.123,
        updated_at=1625819123.123,
    )
    result = hit.to_dict()
    assert "metadata" not in result
    assert "filters" not in result



"""
RedisVL Semantic Session Manager (Deprecated Path)

This module is kept for backward compatibility. Please use `redisvl.extensions.semantic_history` instead.
"""

import warnings

from redisvl.extensions.message_history.semantic_history import SemanticMessageHistory

warnings.warn(
    "Importing from redisvl.extensions.session_manger.semantic_session is deprecated. "
    "SemanticSessionManager has been renamed to SemanticMessageHistory. "
    "Please import SemanticMessageHistory from redisvl.extensions.semantic_history instead.",
    DeprecationWarning,
    stacklevel=2,
)


class SemanticSessionManager(SemanticMessageHistory):
    # keep for backwards compatibility
    pass


__all__ = ["SemanticSessionManager"]



import os
import pathlib

import pytest

from redisvl.schema.fields import TagField, TextField
from redisvl.schema.schema import IndexSchema, StorageType


def get_base_path():
    return pathlib.Path(__file__).parent.resolve()


# Sample data for testing
def create_sample_index_schema():
    sample_fields = [
        {"name": "example_text", "type": "text", "attrs": {"sortable": False}},
        {"name": "example_numeric", "type": "numeric", "attrs": {"sortable": True}},
        {"name": "example_tag", "type": "tag", "attrs": {"sortable": True}},
        {
            "name": "example_vector",
            "type": "vector",
            "attrs": {"dims": 1024, "algorithm": "flat"},
        },
    ]
    return IndexSchema.from_dict({"index": {"name": "test"}, "fields": sample_fields})


# Tests for IndexSchema


def test_initialization_with_default_params():
    """Test basic schema init with defaults."""
    default_schema = IndexSchema.from_dict({"index": {"name": "test"}})
    assert default_schema.version == "0.1.0"
    assert default_schema.index.name == "test"
    assert default_schema.index.prefix == "rvl"  # Default value
    assert default_schema.index.key_separator == ":"  # Default value
    assert default_schema.index.storage_type == StorageType.HASH  # Default value
    assert default_schema.fields == {}  # Default value


def test_initialization_with_custom_params():
    """Test custom schema params."""
    custom_schema = IndexSchema.from_dict(
        {
            "index": {
                "name": "custom_schema",
                "prefix": "custom",
                "key_separator": "|",
                "storage_type": "json",
            }
        }
    )
    assert custom_schema.index.name == "custom_schema"
    assert custom_schema.index.prefix == "custom"
    assert custom_schema.index.key_separator == "|"
    assert custom_schema.index.storage_type == StorageType.JSON


def test_add_field():
    """Test field addition."""
    index_schema = create_sample_index_schema()
    index_schema.add_field({"name": "new_text_field", "type": "text"})
    assert "new_text_field" in index_schema.fields
    assert isinstance(index_schema.fields["new_text_field"], TextField)


def test_add_fields():
    """Test multiple field addition."""
    index_schema = create_sample_index_schema()
    index_schema.add_fields(
        [
            {"name": "new_text_field", "type": "text"},
            {"name": "new_tag_field", "type": "tag"},
        ]
    )
    assert "new_text_field" in index_schema.fields
    assert isinstance(index_schema.fields["new_text_field"], TextField)
    assert "new_tag_field" in index_schema.fields
    assert isinstance(index_schema.fields["new_tag_field"], TagField)


def test_add_duplicate_field():
    """Test adding a duplicate field."""
    index_schema = create_sample_index_schema()
    with pytest.raises(ValueError):
        index_schema.add_field({"name": "example_text", "type": "text"})


def test_remove_field():
    """Test field removal."""
    index_schema = create_sample_index_schema()
    index_schema.remove_field("example_text")
    assert "example_text" not in index_schema.field_names


def test_generate_fields():
    """Test field generation."""
    sample = {"name": "John", "age": 30, "tags": ["test", "test2"]}
    index_schema = IndexSchema.from_dict({"index": {"name": "test"}})
    generated_fields = index_schema.generate_fields(sample)
    expected_field_names = sample.keys()
    for field in generated_fields:
        assert field["name"] in expected_field_names
        assert field["path"] == None


def test_to_dict():
    """Test schema to dict serialization."""
    index_schema = create_sample_index_schema()
    index_dict = index_schema.to_dict()
    assert index_dict["index"]["name"] == "test"
    assert isinstance(index_dict["fields"], list)
    assert len(index_dict["fields"]) == 4 == len(index_schema.fields)


def test_from_dict():
    """Test loading schema from a dictionary."""
    sample_fields = [
        {"name": "example_text", "type": "text", "attrs": {"sortable": False}},
        {"name": "example_numeric", "type": "tag", "attrs": {"sortable": True}},
    ]
    index_schema = IndexSchema.from_dict(
        {
            "index": {
                "name": "example_index",
                "prefix": "ex",
                "key_separator": "|",
                "storage_type": "json",
            },
            "fields": sample_fields,
        }
    )
    assert index_schema.index.name == "example_index"
    assert index_schema.index.key_separator == "|"
    assert index_schema.index.prefix == "ex"
    assert index_schema.index.storage_type == StorageType.JSON
    assert len(index_schema.fields) == 2


def test_hash_index_from_yaml():
    """Test loading from yaml."""
    index_schema = IndexSchema.from_yaml(
        str(get_base_path().joinpath("../../schemas/test_hash_schema.yaml"))
    )
    assert index_schema.index.name == "hash-test"
    assert index_schema.index.prefix == "hash"
    assert index_schema.index.storage_type == StorageType.HASH
    assert len(index_schema.fields) == 2


def test_json_index_from_yaml():
    """Test loading from yaml."""
    index_schema = IndexSchema.from_yaml(
        str(get_base_path().joinpath("../../schemas/test_json_schema.yaml"))
    )
    assert index_schema.index.name == "json-test"
    assert index_schema.index.prefix == "json"
    assert index_schema.index.storage_type == StorageType.JSON
    assert len(index_schema.fields) == 2


def test_to_yaml_and_reload():
    index_schema = create_sample_index_schema()
    index_schema.to_yaml("temp_test.yaml")

    assert os.path.exists("temp_test.yaml")

    new_schema = IndexSchema.from_yaml("temp_test.yaml")
    assert new_schema == index_schema
    assert new_schema.to_dict() == index_schema.to_dict()

    os.remove("temp_test.yaml")


def test_from_yaml_file_not_found():
    """Test loading from yaml with file not found."""
    with pytest.raises(FileNotFoundError):
        IndexSchema.from_yaml("nonexistent_file")


def test_schema_with_index_missing_and_empty_attributes():
    """Test schema creation and operations with INDEXMISSING and INDEXEMPTY attributes."""
    schema_dict = {
        "index": {
            "name": "test-missing-empty",
            "prefix": "test",
            "storage_type": "hash",
        },
        "fields": [
            {
                "name": "title",
                "type": "text",
                "attrs": {"index_missing": True, "index_empty": True, "sortable": True},
            },
            {
                "name": "tags",
                "type": "tag",
                "attrs": {"index_missing": True, "index_empty": True},
            },
            {
                "name": "price",
                "type": "numeric",
                "attrs": {"index_missing": True, "sortable": True},
            },
            {"name": "location", "type": "geo", "attrs": {"index_missing": True}},
            {
                "name": "embedding",
                "type": "vector",
                "attrs": {
                    "algorithm": "flat",
                    "dims": 128,
                    "distance_metric": "cosine",
                    "index_missing": True,
                },
            },
        ],
    }

    # Test schema creation
    schema = IndexSchema.from_dict(schema_dict)

    # Verify field attributes are correctly set
    assert schema.fields["title"].attrs.index_missing == True
    assert schema.fields["title"].attrs.index_empty == True
    assert schema.fields["title"].attrs.sortable == True

    assert schema.fields["tags"].attrs.index_missing == True
    assert schema.fields["tags"].attrs.index_empty == True

    assert schema.fields["price"].attrs.index_missing == True
    assert schema.fields["price"].attrs.sortable == True

    assert schema.fields["location"].attrs.index_missing == True

    assert schema.fields["embedding"].attrs.index_missing == True
    assert schema.fields["embedding"].attrs.dims == 128

    # Test Redis field conversion
    redis_fields = schema.redis_fields
    assert len(redis_fields) == 5

    # Verify all fields can be converted to Redis fields successfully
    for field_name, field in schema.fields.items():
        redis_field = field.as_redis_field()
        assert redis_field.name == field_name


def test_schema_serialization_with_new_attributes():
    """Test schema creation and field attribute handling with INDEXMISSING and INDEXEMPTY attributes."""
    original_schema_dict = {
        "index": {
            "name": "test-serialization",
            "prefix": "set",
            "storage_type": "hash",
        },
        "fields": [
            {
                "name": "description",
                "type": "text",
                "attrs": {"index_missing": True, "index_empty": True, "weight": 2.0},
            },
            {
                "name": "categories",
                "type": "tag",
                "attrs": {"index_missing": True, "index_empty": True, "separator": "|"},
            },
            {"name": "score", "type": "numeric", "attrs": {"index_missing": True}},
            {
                "name": "vector_field",
                "type": "vector",
                "attrs": {
                    "algorithm": "hnsw",
                    "dims": 256,
                    "index_missing": True,
                    "m": 24,
                },
            },
        ],
    }

    # Create schema from dict
    schema = IndexSchema.from_dict(original_schema_dict)

    # Verify field attributes are correctly set after creation
    assert schema.fields["description"].attrs.index_missing == True
    assert schema.fields["description"].attrs.index_empty == True
    assert schema.fields["description"].attrs.weight == 2.0

    assert schema.fields["categories"].attrs.index_missing == True
    assert schema.fields["categories"].attrs.index_empty == True
    assert schema.fields["categories"].attrs.separator == "|"

    assert schema.fields["score"].attrs.index_missing == True

    assert schema.fields["vector_field"].attrs.index_missing == True
    assert schema.fields["vector_field"].attrs.dims == 256
    assert schema.fields["vector_field"].attrs.m == 24

    # Test that Redis field conversion works with new attributes
    for field_name, field in schema.fields.items():
        redis_field = field.as_redis_field()
        assert redis_field.name == field_name

    # Test that the schema has the correct number of fields
    assert len(schema.fields) == 4
    assert schema.index.name == "test-serialization"



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('.'))

print(f"Sphinx is using Python executable at: {sys.executable}", flush=True)
print(f"Python version: {sys.version}", flush=True)

# -- Project information -----------------------------------------------------

from redisvl import __version__

project = 'RedisVL'
copyright = '2024, Redis Inc.'
author = 'Redis Applied AI'
version = __version__

# The full version, including alpha/beta/rc tags
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.imgmath',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    "sphinx_design",
    "sphinx_copybutton",
    "_extension.gallery_directive",
    "myst_nb",
    "sphinx_favicon"
]


# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', "**.ipynb_checkpoints"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.
html_theme = "sphinx_book_theme"

# Pygments syntax highlighting style
pygments_style = "friendly"
pygments_dark_style = "monokai"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = ["css/custom.css"]
html_title = "RedisVL"
html_logo = "_static/Redis_Favicon_32x32_Red.png"
html_favicon = "_static/Redis_Favicon_32x32_Red.png"

html_context = {
    "github_user": "redis",
    "github_repo": "redis-vl-python",
    "github_version": "main",
    "doc_path": "docs",
    "default_mode": "auto",
}

# This allows us to use ::: to denote directives, useful for admonitions
myst_enable_extensions = ["colon_fence"]
myst_heading_anchors = 3

# Sphinx Book Theme options
html_theme_options = {
    "repository_url": "https://github.com/redis/redis-vl-python",
    "use_repository_button": True,
    "use_edit_page_button": True,
    "use_source_button": True,
    "use_issues_button": True,
    "use_download_button": True,
    "use_fullscreen_button": True,
    "repository_branch": "main",
    "path_to_docs": "docs",
    "show_navbar_depth": 2,
    "navigation_depth": 4,
    "show_toc_level": 3,
    "home_page_in_toc": True,
    "logo": {
        "text": "Redis Vector Library",
        "image_light": "_static/Redis_Favicon_32x32_Red.png",
        "image_dark": "_static/Redis_Favicon_32x32_Red.png",
    },
}


autoclass_content = 'both'
add_module_names = False

nb_execution_mode = "off"

# -- Options for autosummary/autodoc output ------------------------------------
autosummary_generate = True
autodoc_typehints = "description"
autodoc_member_order = "groupwise"

# -- Options for autoapi -------------------------------------------------------
autoapi_type = "python"
autoapi_dirs = ["../src/redisvl"]
autoapi_keep_files = True
autoapi_root = "api"
autoapi_member_order = "groupwise"


# -- favicon options ---------------------------------------------------------

# see https://sphinx-favicon.readthedocs.io for more information about the
# sphinx-favicon extension

favicons = [
    # generic icons compatible with most browsers
    "Redis_Favicon_32x32_Red.png",
    "Redis_Favicon_16x16_Red.png",
    "Redis_Favicon_144x144_Red.png",
]



"""Integration tests for role filtering in get_recent() method."""

import pytest

from redisvl.extensions.message_history import MessageHistory, SemanticMessageHistory


class TestMessageHistoryRoleFilter:
    """Test role filtering functionality in MessageHistory with real Redis."""

    def test_get_recent_single_role_system(self, redis_url):
        """Test get_recent with role='system' returns only system messages."""
        history = MessageHistory("test_role_system", redis_url=redis_url)

        # Clear any existing data
        history.clear()

        # Add various messages with different roles
        history.add_messages(
            [
                {"role": "system", "content": "System initialization"},
                {"role": "user", "content": "Hello"},
                {"role": "llm", "content": "Hi there"},
                {"role": "system", "content": "System configuration updated"},
                {
                    "role": "tool",
                    "content": "Function executed",
                    "tool_call_id": "call1",
                },
            ]
        )

        # Get only system messages
        result = history.get_recent(role="system", top_k=10)

        assert len(result) == 2
        assert all(msg["role"] == "system" for msg in result)
        assert result[0]["content"] == "System initialization"
        assert result[1]["content"] == "System configuration updated"

        # Cleanup
        history.delete()

    def test_get_recent_single_role_user(self, redis_url):
        """Test get_recent with role='user' returns only user messages."""
        history = MessageHistory("test_role_user", redis_url=redis_url)
        history.clear()

        history.add_messages(
            [
                {"role": "system", "content": "Welcome"},
                {"role": "user", "content": "First question"},
                {"role": "llm", "content": "First answer"},
                {"role": "user", "content": "Second question"},
                {"role": "user", "content": "Third question"},
            ]
        )

        result = history.get_recent(role="user", top_k=10)

        assert len(result) == 3
        assert all(msg["role"] == "user" for msg in result)
        assert result[0]["content"] == "First question"
        assert result[2]["content"] == "Third question"

        history.delete()

    def test_get_recent_single_role_llm(self, redis_url):
        """Test get_recent with role='llm' returns only llm messages."""
        history = MessageHistory("test_role_llm", redis_url=redis_url)
        history.clear()

        history.add_messages(
            [
                {"role": "user", "content": "Question 1"},
                {"role": "llm", "content": "Answer 1"},
                {"role": "user", "content": "Question 2"},
                {"role": "llm", "content": "Answer 2"},
                {"role": "system", "content": "System note"},
            ]
        )

        result = history.get_recent(role="llm", top_k=10)

        assert len(result) == 2
        assert all(msg["role"] == "llm" for msg in result)
        assert result[0]["content"] == "Answer 1"
        assert result[1]["content"] == "Answer 2"

        history.delete()

    def test_get_recent_single_role_tool(self, redis_url):
        """Test get_recent with role='tool' returns only tool messages."""
        history = MessageHistory("test_role_tool", redis_url=redis_url)
        history.clear()

        history.add_messages(
            [
                {"role": "user", "content": "Run function"},
                {
                    "role": "tool",
                    "content": "Function result 1",
                    "tool_call_id": "call1",
                },
                {"role": "llm", "content": "Processing"},
                {
                    "role": "tool",
                    "content": "Function result 2",
                    "tool_call_id": "call2",
                },
            ]
        )

        result = history.get_recent(role="tool", top_k=10)

        assert len(result) == 2
        assert all(msg["role"] == "tool" for msg in result)
        assert all("tool_call_id" in msg for msg in result)

        history.delete()

    def test_get_recent_multiple_roles(self, redis_url):
        """Test get_recent with multiple roles returns matching messages."""
        history = MessageHistory("test_multi_roles", redis_url=redis_url)
        history.clear()

        history.add_messages(
            [
                {"role": "system", "content": "System message"},
                {"role": "user", "content": "User message"},
                {"role": "llm", "content": "LLM message"},
                {"role": "tool", "content": "Tool message", "tool_call_id": "call1"},
            ]
        )

        # Get system and user messages only
        result = history.get_recent(role=["system", "user"], top_k=10)

        assert len(result) == 2
        assert all(msg["role"] in ["system", "user"] for msg in result)
        assert result[0]["content"] == "System message"
        assert result[1]["content"] == "User message"

        history.delete()

    def test_get_recent_no_role_filter_backward_compatibility(self, redis_url):
        """Test get_recent with role=None returns all messages (backward compatibility)."""
        history = MessageHistory("test_no_filter", redis_url=redis_url)
        history.clear()

        history.add_messages(
            [
                {"role": "system", "content": "System"},
                {"role": "user", "content": "User"},
                {"role": "llm", "content": "LLM"},
                {"role": "tool", "content": "Tool", "tool_call_id": "call1"},
            ]
        )

        # No role filter - should return all messages
        result = history.get_recent(role=None, top_k=10)

        assert len(result) == 4
        roles = {msg["role"] for msg in result}
        assert roles == {"system", "user", "llm", "tool"}

        history.delete()

    def test_get_recent_invalid_role_raises_error(self, redis_url):
        """Test get_recent with invalid role raises ValueError."""
        history = MessageHistory("test_invalid", redis_url=redis_url)

        with pytest.raises(ValueError, match="Invalid role"):
            history.get_recent(role="invalid_role")

        history.delete()

    def test_get_recent_invalid_role_in_list_raises_error(self, redis_url):
        """Test get_recent with invalid role in list raises ValueError."""
        history = MessageHistory("test_invalid_list", redis_url=redis_url)

        with pytest.raises(ValueError, match="Invalid role"):
            history.get_recent(role=["system", "invalid_role"])

        history.delete()

    def test_get_recent_empty_role_list_raises_error(self, redis_url):
        """Test get_recent with empty role list raises ValueError."""
        history = MessageHistory("test_empty_list", redis_url=redis_url)

        with pytest.raises(ValueError, match="roles cannot be empty"):
            history.get_recent(role=[])

        history.delete()

    def test_get_recent_role_with_other_parameters(self, redis_url):
        """Test role filter works with other parameters like top_k."""
        history = MessageHistory("test_with_params", redis_url=redis_url)
        history.clear()

        # Add many system messages
        for i in range(5):
            history.add_message({"role": "system", "content": f"System message {i}"})

        # Add other messages
        history.add_message({"role": "user", "content": "User message"})
        history.add_message({"role": "llm", "content": "LLM message"})

        # Get only 2 most recent system messages
        result = history.get_recent(role="system", top_k=2)

        assert len(result) == 2
        assert all(msg["role"] == "system" for msg in result)
        # Should get most recent ones
        assert result[0]["content"] == "System message 3"
        assert result[1]["content"] == "System message 4"

        history.delete()

    def test_get_recent_role_with_session_tag(self, redis_url):
        """Test role filter works with session_tag parameter."""
        history = MessageHistory("test_session", redis_url=redis_url)
        history.clear()

        # Add messages with different session tags
        history.add_messages(
            [
                {"role": "system", "content": "System for session1"},
                {"role": "user", "content": "User for session1"},
            ],
            session_tag="session1",
        )

        history.add_messages(
            [
                {"role": "system", "content": "System for session2"},
                {"role": "llm", "content": "LLM for session2"},
            ],
            session_tag="session2",
        )

        # Get system messages from session2 only
        result = history.get_recent(role="system", session_tag="session2", top_k=10)

        assert len(result) == 1
        assert result[0]["role"] == "system"
        assert result[0]["content"] == "System for session2"

        history.delete()

    def test_get_recent_role_with_raw_output(self, redis_url):
        """Test role filter works with raw=True."""
        history = MessageHistory("test_raw", redis_url=redis_url)
        history.clear()

        history.add_message({"role": "system", "content": "System message"})

        result = history.get_recent(role="system", raw=True, top_k=10)

        assert len(result) == 1
        assert result[0]["role"] == "system"
        # Raw should include additional metadata
        assert "entry_id" in result[0]
        assert "timestamp" in result[0]
        assert "session_tag" in result[0]

        history.delete()


class TestSemanticMessageHistoryRoleFilter:
    """Test role filtering functionality in SemanticMessageHistory with real Redis."""

    def test_semantic_get_recent_with_role(self, redis_url):
        """Test SemanticMessageHistory get_recent with role filter."""
        history = SemanticMessageHistory(
            "test_semantic_recent", redis_url=redis_url, overwrite=True
        )
        history.clear()

        history.add_messages(
            [
                {"role": "system", "content": "System prompt about configuration"},
                {"role": "user", "content": "User question"},
                {"role": "llm", "content": "Assistant response"},
            ]
        )

        result = history.get_recent(role="system", top_k=10)

        assert len(result) == 1
        assert result[0]["role"] == "system"
        assert "configuration" in result[0]["content"]

        history.delete()

    def test_semantic_get_relevant_with_role(self, redis_url):
        """Test SemanticMessageHistory get_relevant with role filter."""
        history = SemanticMessageHistory(
            "test_semantic_relevant", redis_url=redis_url, overwrite=True
        )
        history.clear()

        history.add_messages(
            [
                {
                    "role": "system",
                    "content": "System instructions about fruits and vegetables",
                },
                {"role": "user", "content": "Tell me about apples"},
                {"role": "llm", "content": "Apples are a type of fruit"},
                {"role": "user", "content": "What about cars?"},
                {"role": "llm", "content": "Cars are vehicles for transportation"},
            ]
        )

        # Search for fruit-related messages but only from system role
        result = history.get_relevant("fruits", role="system", top_k=10)

        if result:  # Semantic search might not find exact matches
            assert all(msg["role"] == "system" for msg in result)

        # Search for fruit-related messages from user role
        result = history.get_relevant("apples", role="user", top_k=10)

        if result:
            assert all(msg["role"] == "user" for msg in result)

        history.delete()


class TestRoleValidation:
    """Test role validation logic."""

    def test_valid_roles_accepted(self, redis_url):
        """Test that all valid roles are accepted."""
        valid_roles = ["system", "user", "llm", "tool"]
        history = MessageHistory("test_valid_roles", redis_url=redis_url)
        history.clear()

        # Add messages with all valid roles
        for role in valid_roles:
            if role == "tool":
                history.add_message(
                    {
                        "role": role,
                        "content": f"{role} message",
                        "tool_call_id": "call1",
                    }
                )
            else:
                history.add_message({"role": role, "content": f"{role} message"})

        # Test each valid role works
        for role in valid_roles:
            result = history.get_recent(role=role, top_k=10)
            assert len(result) >= 1
            assert all(msg["role"] == role for msg in result)

        history.delete()

    def test_case_sensitive_roles(self, redis_url):
        """Test that role validation is case sensitive."""
        history = MessageHistory("test_case", redis_url=redis_url)

        with pytest.raises(ValueError):
            history.get_recent(role="SYSTEM")  # uppercase should fail

        with pytest.raises(ValueError):
            history.get_recent(role="User")  # mixed case should fail

        history.delete()



from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from redisvl.extensions.constants import (
    CACHE_VECTOR_FIELD_NAME,
    INSERTED_AT_FIELD_NAME,
    PROMPT_FIELD_NAME,
    RESPONSE_FIELD_NAME,
    UPDATED_AT_FIELD_NAME,
)
from redisvl.redis.utils import array_to_buffer, hashify
from redisvl.schema import IndexSchema
from redisvl.utils.utils import current_timestamp, deserialize, serialize


class CacheEntry(BaseModel):
    """A single cache entry in Redis"""

    entry_id: Optional[str] = Field(default=None)
    """Cache entry identifier"""
    prompt: str
    """Input prompt or question cached in Redis"""
    response: str
    """Response or answer to the question, cached in Redis"""
    prompt_vector: List[float]
    """Text embedding representation of the prompt"""
    inserted_at: float = Field(default_factory=current_timestamp)
    """Timestamp of when the entry was added to the cache"""
    updated_at: float = Field(default_factory=current_timestamp)
    """Timestamp of when the entry was updated in the cache"""
    metadata: Optional[Dict[str, Any]] = Field(default=None)
    """Optional metadata stored on the cache entry"""
    filters: Optional[Dict[str, Any]] = Field(default=None)
    """Optional filter data stored on the cache entry for customizing retrieval"""

    @model_validator(mode="before")
    @classmethod
    def generate_id(cls, values):
        # Ensure entry_id is set
        if not values.get("entry_id"):
            values["entry_id"] = hashify(values["prompt"], values.get("filters"))
        return values

    @field_validator("metadata")
    @classmethod
    def non_empty_metadata(cls, v):
        if v is not None and not isinstance(v, dict):
            raise TypeError("Metadata must be a dictionary.")
        return v

    def to_dict(self, dtype: str) -> Dict:
        data = self.model_dump(exclude_none=True)
        data["prompt_vector"] = array_to_buffer(self.prompt_vector, dtype)
        if self.metadata is not None:
            data["metadata"] = serialize(self.metadata)
        if self.filters is not None:
            data.update(self.filters)
            del data["filters"]
        return data


class CacheHit(BaseModel):
    """A cache hit based on some input query"""

    entry_id: str
    """Cache entry identifier"""
    prompt: str
    """Input prompt or question cached in Redis"""
    response: str
    """Response or answer to the question, cached in Redis"""
    vector_distance: float
    """The semantic distance between the query vector and the stored prompt vector"""
    inserted_at: float
    """Timestamp of when the entry was added to the cache"""
    updated_at: float
    """Timestamp of when the entry was updated in the cache"""
    metadata: Optional[Dict[str, Any]] = Field(default=None)
    """Optional metadata stored on the cache entry"""
    filters: Optional[Dict[str, Any]] = Field(default=None)
    """Optional filter data stored on the cache entry for customizing retrieval"""

    # Allow extra fields to simplify handling filters
    model_config = ConfigDict(extra="allow")

    @model_validator(mode="before")
    @classmethod
    def validate_cache_hit(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        # Deserialize metadata if necessary
        if "metadata" in values and isinstance(values["metadata"], str):
            values["metadata"] = deserialize(values["metadata"])

        # Collect any extra fields and store them as filters
        extra_data = values.pop("__pydantic_extra__", {}) or {}
        if extra_data:
            current_filters = values.get("filters") or {}
            if not isinstance(current_filters, dict):
                current_filters = {}
            current_filters.update(extra_data)
            values["filters"] = current_filters

        return values

    def to_dict(self) -> Dict[str, Any]:
        """Convert this model to a dictionary, merging filters into the result."""
        data = self.model_dump(exclude_none=True)
        if data.get("filters"):
            data.update(data["filters"])
            del data["filters"]
        return data


class SemanticCacheIndexSchema(IndexSchema):

    @classmethod
    def from_params(cls, name: str, prefix: str, vector_dims: int, dtype: str):
        return cls(
            index={"name": name, "prefix": prefix},  # type: ignore
            fields=[  # type: ignore
                {"name": PROMPT_FIELD_NAME, "type": "text"},
                {"name": RESPONSE_FIELD_NAME, "type": "text"},
                {"name": INSERTED_AT_FIELD_NAME, "type": "numeric"},
                {"name": UPDATED_AT_FIELD_NAME, "type": "numeric"},
                {
                    "name": CACHE_VECTOR_FIELD_NAME,
                    "type": "vector",
                    "attrs": {
                        "dims": vector_dims,
                        "datatype": dtype,
                        "distance_metric": "cosine",
                        "algorithm": "flat",
                    },
                },
            ],
        )



"""
RedisVL Exception Classes

This module defines all custom exceptions used throughout the RedisVL library.
"""


class RedisVLError(Exception):
    """Base exception for all RedisVL errors."""

    pass


class RedisSearchError(RedisVLError):
    """Error raised for Redis Search specific operations."""

    pass


class SchemaValidationError(RedisVLError):
    """Error when validating data against a schema."""

    def __init__(self, message, index=None):
        # Only add index prefix if the message doesn't already contain detailed validation info
        if index is not None and not message.startswith("Schema validation failed"):
            message = f"Validation failed for object at index {index}: {message}"
        super().__init__(message)


class QueryValidationError(RedisVLError):
    """Error when validating a query."""

    pass


class RedisModuleVersionError(RedisVLError):
    """Error when Redis or module versions are incompatible with requested features."""

    @classmethod
    def for_svs_vamana(cls, min_redis_version: str):
        """Create error for unsupported SVS-VAMANA.

        Args:
            min_redis_version: Minimum required Redis version

        Returns:
            RedisModuleVersionError with formatted message
        """
        message = (
            f"SVS-VAMANA requires Redis >= {min_redis_version} with RediSearch >= 2.8.10. "
            f"Options: 1) Upgrade Redis Stack, "
            f"2) Use algorithm='hnsw' or 'flat', "
            f"3) Remove compression parameters"
        )
        return cls(message)



"""Unit tests for stopwords support in IndexSchema."""

import tempfile

import yaml

from redisvl.schema import IndexSchema


def test_index_schema_stopwords_none_default():
    """Test IndexSchema with no stopwords specified (default behavior)."""
    schema_dict = {
        "index": {
            "name": "test_index",
            "prefix": "test",
            "storage_type": "hash",
        },
        "fields": [
            {"name": "title", "type": "text"},
        ],
    }

    schema = IndexSchema.from_dict(schema_dict)

    assert schema.index.name == "test_index"
    assert schema.index.stopwords is None  # Default


def test_index_schema_stopwords_disabled():
    """Test IndexSchema with stopwords disabled (STOPWORDS 0)."""
    schema_dict = {
        "index": {
            "name": "test_index",
            "prefix": "test",
            "storage_type": "hash",
            "stopwords": [],  # Empty list = STOPWORDS 0
        },
        "fields": [
            {"name": "title", "type": "text"},
        ],
    }

    schema = IndexSchema.from_dict(schema_dict)

    assert schema.index.name == "test_index"
    assert schema.index.stopwords == []


def test_index_schema_custom_stopwords():
    """Test IndexSchema with custom stopwords list."""
    schema_dict = {
        "index": {
            "name": "test_index",
            "prefix": "test",
            "storage_type": "hash",
            "stopwords": ["the", "a", "an"],
        },
        "fields": [
            {"name": "title", "type": "text"},
        ],
    }

    schema = IndexSchema.from_dict(schema_dict)

    assert schema.index.name == "test_index"
    assert schema.index.stopwords == ["the", "a", "an"]


def test_index_schema_stopwords_from_yaml_disabled():
    """Test IndexSchema from YAML with stopwords disabled."""
    yaml_content = """
version: '0.1.0'

index:
    name: test_yaml_index
    prefix: test_yaml
    storage_type: hash
    stopwords: []

fields:
    - name: title
      type: text
"""

    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
        f.write(yaml_content)
        yaml_path = f.name

    try:
        schema = IndexSchema.from_yaml(yaml_path)
        assert schema.index.name == "test_yaml_index"
        assert schema.index.stopwords == []
    finally:
        import os

        os.unlink(yaml_path)


def test_index_schema_stopwords_from_yaml_custom():
    """Test IndexSchema from YAML with custom stopwords."""
    yaml_content = """
version: '0.1.0'

index:
    name: test_yaml_index
    prefix: test_yaml
    storage_type: hash
    stopwords:
        - the
        - a
        - an

fields:
    - name: title
      type: text
"""

    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
        f.write(yaml_content)
        yaml_path = f.name

    try:
        schema = IndexSchema.from_yaml(yaml_path)
        assert schema.index.name == "test_yaml_index"
        assert schema.index.stopwords == ["the", "a", "an"]
    finally:
        import os

        os.unlink(yaml_path)


def test_index_schema_to_dict_preserves_stopwords():
    """Test that to_dict() preserves stopwords configuration."""
    schema_dict = {
        "index": {
            "name": "test_index",
            "prefix": "test",
            "storage_type": "hash",
            "stopwords": ["the", "a"],
        },
        "fields": [
            {"name": "title", "type": "text"},
        ],
    }

    schema = IndexSchema.from_dict(schema_dict)
    result_dict = schema.to_dict()

    assert result_dict["index"]["stopwords"] == ["the", "a"]


def test_index_schema_to_dict_omits_none_stopwords():
    """Test that to_dict() omits stopwords when None (default)."""
    schema_dict = {
        "index": {
            "name": "test_index",
            "prefix": "test",
            "storage_type": "hash",
        },
        "fields": [
            {"name": "title", "type": "text"},
        ],
    }

    schema = IndexSchema.from_dict(schema_dict)
    result_dict = schema.to_dict()

    # stopwords should not be in the dict when None (default behavior)
    assert "stopwords" not in result_dict["index"]


def test_index_schema_to_yaml_preserves_stopwords():
    """Test that to_yaml() preserves stopwords configuration."""
    schema_dict = {
        "index": {
            "name": "test_index",
            "prefix": "test",
            "storage_type": "hash",
            "stopwords": [],  # STOPWORDS 0
        },
        "fields": [
            {"name": "title", "type": "text"},
        ],
    }

    schema = IndexSchema.from_dict(schema_dict)

    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
        yaml_path = f.name

    try:
        schema.to_yaml(yaml_path)

        # Read back and verify
        with open(yaml_path, "r") as f:
            yaml_data = yaml.safe_load(f)

        assert yaml_data["index"]["stopwords"] == []
    finally:
        import os

        os.unlink(yaml_path)



import warnings
from typing import Any, Dict, List, Optional, Set, Union

from pydantic import BaseModel, Field, field_validator, model_validator
from redis.commands.search.aggregation import AggregateRequest, Desc
from typing_extensions import Self

from redisvl.query.filter import FilterExpression
from redisvl.redis.utils import array_to_buffer
from redisvl.schema.fields import VectorDataType
from redisvl.utils.full_text_query_helper import FullTextQueryHelper
from redisvl.utils.utils import lazy_import

nltk = lazy_import("nltk")
nltk_stopwords = lazy_import("nltk.corpus.stopwords")


class Vector(BaseModel):
    """
    Simple object containing the necessary arguments to perform a multi vector query.

    Args:
    vector: The vector values as a list of floats or bytes
    field_name: The name of the vector field to search
    dtype: The data type of the vector (default: "float32")
    weight: The weight for this vector in the combined score (default: 1.0)
    max_distance: The maximum distance for vector range search (default: 2.0, range: [0.0, 2.0])
    """

    vector: Union[List[float], bytes]
    field_name: str
    dtype: str = "float32"
    weight: float = 1.0
    max_distance: float = Field(default=2.0, ge=0.0, le=2.0)

    @field_validator("dtype")
    @classmethod
    def validate_dtype(cls, dtype: str) -> str:
        try:
            VectorDataType(dtype.upper())
        except ValueError:
            raise ValueError(
                f"Invalid data type: {dtype}. Supported types are: {[t.lower() for t in VectorDataType]}"
            )
        return dtype

    @field_validator("max_distance")
    @classmethod
    def validate_max_distance(cls, max_distance: float) -> float:
        if not isinstance(max_distance, (float, int)):
            raise ValueError("max_distance must be a value between 0.0 and 2.0")
        if max_distance < 0.0 or max_distance > 2.0:
            raise ValueError("max_distance must be a value between 0.0 and 2.0")
        return max_distance

    @model_validator(mode="after")
    def validate_vector(self) -> Self:
        """If the vector passed in is an array of float convert it to a byte string."""
        if isinstance(self.vector, bytes):
            return self
        self.vector = array_to_buffer(self.vector, self.dtype)
        return self


class AggregationQuery(AggregateRequest):
    """
    Base class for aggregation queries used to create aggregation queries for Redis.
    """

    def __init__(self, query_string):
        super().__init__(query_string)


class AggregateHybridQuery(AggregationQuery):
    """
    AggregateHybridQuery combines text and vector search in Redis.
    It allows you to perform a hybrid search using both text and vector similarity.
    It scores documents based on a weighted combination of text and vector similarity.

    .. code-block:: python

        from redisvl.query import AggregateHybridQuery
        from redisvl.index import SearchIndex

        index = SearchIndex.from_yaml("path/to/index.yaml")

        query = AggregateHybridQuery(
            text="example text",
            text_field_name="text_field",
            vector=[0.1, 0.2, 0.3],
            vector_field_name="vector_field",
            text_scorer="BM25STD",
            filter_expression=None,
            alpha=0.7,
            dtype="float32",
            num_results=10,
            return_fields=["field1", "field2"],
            stopwords="english",
            dialect=2,
        )

        results = index.query(query)

    """

    DISTANCE_ID: str = "vector_distance"
    VECTOR_PARAM: str = "vector"

    def __init__(
        self,
        text: str,
        text_field_name: str,
        vector: Union[bytes, List[float]],
        vector_field_name: str,
        text_scorer: str = "BM25STD",
        filter_expression: Optional[Union[str, FilterExpression]] = None,
        alpha: float = 0.7,
        dtype: str = "float32",
        num_results: int = 10,
        return_fields: Optional[List[str]] = None,
        stopwords: Optional[Union[str, Set[str]]] = "english",
        dialect: int = 2,
        text_weights: Optional[Dict[str, float]] = None,
    ):
        """
        Instantiates a AggregateHybridQuery object.

        Args:
            text (str): The text to search for.
            text_field_name (str): The text field name to search in.
            vector (Union[bytes, List[float]]): The vector to perform vector similarity search.
            vector_field_name (str): The vector field name to search in.
            text_scorer (str, optional): The text scorer to use. Options are {TFIDF, TFIDF.DOCNORM,
                BM25, DISMAX, DOCSCORE, BM25STD}. Defaults to "BM25STD".
            filter_expression (Optional[FilterExpression], optional): The filter expression to use.
                Defaults to None.
            alpha (float, optional): The weight of the vector similarity. Documents will be scored
                as: hybrid_score = (alpha) * vector_score + (1-alpha) * text_score.
                Defaults to 0.7.
            dtype (str, optional): The data type of the vector. Defaults to "float32".
            num_results (int, optional): The number of results to return. Defaults to 10.
            return_fields (Optional[List[str]], optional): The fields to return. Defaults to None.
            stopwords (Optional[Union[str, Set[str]]], optional): The stopwords to remove from the
                provided text prior to search-use. If a string such as "english" "german" is
                provided then a default set of stopwords for that language will be used. if a list,
                set, or tuple of strings is provided then those will be used as stopwords.
                Defaults to "english". if set to "None" then no stopwords will be removed.

                Note: This parameter controls query-time stopword filtering (client-side).
                For index-level stopwords configuration (server-side), see IndexInfo.stopwords.
                Using query-time stopwords with index-level STOPWORDS 0 is counterproductive.
            dialect (int, optional): The Redis dialect version. Defaults to 2.
            text_weights (Optional[Dict[str, float]]): The importance weighting of individual words
                within the query text. Defaults to None, as no modifications will be made to the
                text_scorer score.

        Note:
            AggregateHybridQuery uses FT.AGGREGATE commands which do NOT support runtime
            parameters. For runtime parameter support (ef_runtime, search_window_size, etc.),
            use VectorQuery or VectorRangeQuery which use FT.SEARCH commands.

        Raises:
            ValueError: If the text string is empty, or if the text string becomes empty after
                stopwords are removed.
            TypeError: If the stopwords are not a set, list, or tuple of strings.
        """

        if not text.strip():
            raise ValueError("text string cannot be empty")

        self._text = text
        self._text_field = text_field_name
        self._vector = vector
        self._vector_field = vector_field_name
        self._filter_expression = filter_expression
        self._alpha = alpha
        self._dtype = dtype
        self._num_results = num_results

        self._ft_helper = FullTextQueryHelper(
            stopwords=stopwords,
            text_weights=text_weights,
        )

        query_string = self._build_query_string()
        super().__init__(query_string)

        self.scorer(text_scorer)
        self.add_scores()
        self.apply(
            vector_similarity=f"(2 - @{self.DISTANCE_ID})/2", text_score="@__score"
        )
        self.apply(hybrid_score=f"{1-alpha}*@text_score + {alpha}*@vector_similarity")
        self.sort_by(Desc("@hybrid_score"), max=num_results)  # type: ignore
        self.dialect(dialect)
        if return_fields:
            self.load(*return_fields)  # type: ignore[arg-type]

    @property
    def params(self) -> Dict[str, Any]:
        """Return the parameters for the aggregation.

        Returns:
            Dict[str, Any]: The parameters for the aggregation.
        """
        if isinstance(self._vector, list):
            vector = array_to_buffer(self._vector, dtype=self._dtype)
        else:
            vector = self._vector

        params: Dict[str, Any] = {self.VECTOR_PARAM: vector}

        return params

    @property
    def stopwords(self) -> Set[str]:
        """Return the stopwords used in the query.
        Returns:
            Set[str]: The stopwords used in the query.
        """
        return self._ft_helper.stopwords

    @property
    def text_weights(self) -> Dict[str, float]:
        """Get the text weights.

        Returns:
            Dictionary of word:weight mappings.
        """
        return self._ft_helper.text_weights

    def set_text_weights(self, weights: Dict[str, float]):
        """Set or update the text weights for the query.

        Args:
            weights: Dictionary of word:weight mappings
        """
        self._ft_helper.set_text_weights(weights)
        self._query = self._build_query_string()

    def _build_query_string(self) -> str:
        """Build the full query string for text search with optional filtering."""
        text = self._ft_helper.build_query_string(
            self._text, self._text_field, self._filter_expression
        )

        # Build KNN query
        knn_query = (
            f"KNN {self._num_results} @{self._vector_field} ${self.VECTOR_PARAM}"
        )

        # Add distance field alias
        knn_query += f" AS {self.DISTANCE_ID}"

        return f"{text}=>[{knn_query}]"

    def __str__(self) -> str:
        """Return the string representation of the query."""
        return " ".join([str(x) for x in self.build_args()])


class MultiVectorQuery(AggregationQuery):
    """
    MultiVectorQuery allows for search over multiple vector fields in a document simultaneously.
    The final score will be a weighted combination of the individual vector similarity scores
    following the formula:

    score = (w_1 * score_1 + w_2 * score_2 + w_3 * score_3 + ... )

    Vectors may be of different size and datatype, but must be indexed using the 'cosine' distance_metric.

    .. code-block:: python

        from redisvl.query import MultiVectorQuery, Vector
        from redisvl.index import SearchIndex

        index = SearchIndex.from_yaml("path/to/index.yaml")

        vector_1 = Vector(
            vector=[0.1, 0.2, 0.3],
            field_name="text_vector",
            dtype="float32",
            weight=0.7,
        )
        vector_2 = Vector(
            vector=[0.5, 0.5],
            field_name="image_vector",
            dtype="bfloat16",
            weight=0.2,
        )
        vector_3 = Vector(
            vector=[0.1, 0.2, 0.3],
            field_name="text_vector",
            dtype="float64",
            weight=0.5,
        )

        query = MultiVectorQuery(
            vectors=[vector_1, vector_2, vector_3],
            filter_expression=None,
            num_results=10,
            return_fields=["field1", "field2"],
            dialect=2,
        )

        results = index.query(query)
    """

    _vectors: List[Vector]

    def __init__(
        self,
        vectors: Union[Vector, List[Vector]],
        return_fields: Optional[List[str]] = None,
        filter_expression: Optional[Union[str, FilterExpression]] = None,
        num_results: int = 10,
        dialect: int = 2,
    ):
        """
        Instantiates a MultiVectorQuery object.

        Args:
            vectors (Union[Vector, List[Vector]]): The Vectors to perform vector similarity search.
            return_fields (Optional[List[str]], optional): The fields to return. Defaults to None.
            filter_expression (Optional[Union[str, FilterExpression]]): The filter expression to use.
                Defaults to None.
            num_results (int, optional): The number of results to return. Defaults to 10.
            dialect (int, optional): The Redis dialect version. Defaults to 2.
        """

        self._filter_expression = filter_expression
        self._num_results = num_results

        if isinstance(vectors, Vector):
            self._vectors = [vectors]
        else:
            self._vectors = vectors  # type: ignore

        if not all([isinstance(v, Vector) for v in self._vectors]):
            raise TypeError(
                "vector argument must be a Vector object or list of Vector objects."
            )

        query_string = self._build_query_string()
        super().__init__(query_string)

        # calculate the respective vector similarities
        for i in range(len(self._vectors)):
            self.apply(**{f"score_{i}": f"(2 - @distance_{i})/2"})

        # construct the scoring string based on the vector similarity scores and weights
        combined_scores = []
        for i, w in enumerate([v.weight for v in self._vectors]):
            combined_scores.append(f"@score_{i} * {w}")
        combined_score_string = " + ".join(combined_scores)

        self.apply(combined_score=combined_score_string)

        self.sort_by(Desc("@combined_score"), max=num_results)  # type: ignore
        self.dialect(dialect)
        if return_fields:
            self.load(*return_fields)  # type: ignore[arg-type]

    @property
    def params(self) -> Dict[str, Any]:
        """Return the parameters for the aggregation.

        Returns:
            Dict[str, Any]: The parameters for the aggregation.
        """
        params = {}
        for i, v in enumerate(self._vectors):
            params[f"vector_{i}"] = v.vector
        return params

    def _build_query_string(self) -> str:
        """Build the full query string for text search with optional filtering."""

        # base KNN query
        range_queries = []
        for i, (vector, field, max_dist) in enumerate(
            [(v.vector, v.field_name, v.max_distance) for v in self._vectors]
        ):
            range_queries.append(
                f"@{field}:[VECTOR_RANGE {max_dist} $vector_{i}]=>{{$YIELD_DISTANCE_AS: distance_{i}}}"
            )

        range_query = " AND ".join(range_queries)

        filter_expression = self._filter_expression
        if isinstance(self._filter_expression, FilterExpression):
            filter_expression = str(self._filter_expression)

        if filter_expression:
            return f"({range_query}) AND ({filter_expression})"
        else:
            return f"{range_query}"

    def __str__(self) -> str:
        """Return the string representation of the query."""
        return " ".join([str(x) for x in self.build_args()])



"""
Integration tests for SVS-VAMANA vector indexing.

These tests require:
- Redis >= 8.2.0
- RediSearch >= 2.8.10
- Environment variable: REDISVL_TEST_SVS=1

Run with:
    REDISVL_TEST_SVS=1 pytest tests/integration/test_svs_integration.py

To run with Redis 8.2+ in Docker:
    docker run -d -p 6379:6379 redis/redis-stack-server:8.2.2-v0
    REDISVL_TEST_SVS=1 pytest tests/integration/test_svs_integration.py

Or set the Redis image for the test suite:
    REDIS_IMAGE=redis/redis-stack-server:8.2.2-v0 REDISVL_TEST_SVS=1 pytest tests/integration/test_svs_integration.py
"""

import os

import numpy as np
import pytest

from redisvl.exceptions import RedisModuleVersionError
from redisvl.index import SearchIndex
from redisvl.query import VectorQuery
from redisvl.redis.connection import supports_svs
from redisvl.redis.utils import array_to_buffer
from redisvl.schema import IndexSchema
from redisvl.utils import CompressionAdvisor

# Skip all tests in this module if REDISVL_TEST_SVS is not set
pytestmark = pytest.mark.skipif(
    not os.getenv("REDISVL_TEST_SVS"),
    reason="SVS tests require REDISVL_TEST_SVS=1 and Redis 8.2+",
)


@pytest.fixture
def svs_schema_lvq(worker_id):
    """Create SVS-VAMANA schema with LVQ compression."""
    return IndexSchema.from_dict(
        {
            "index": {
                "name": f"svs_lvq_{worker_id}",
                "prefix": f"svs_lvq_{worker_id}",
            },
            "fields": [
                {"name": "id", "type": "tag"},
                {"name": "content", "type": "text"},
                {
                    "name": "embedding",
                    "type": "vector",
                    "attrs": {
                        "dims": 384,
                        "algorithm": "svs-vamana",
                        "datatype": "float32",
                        "distance_metric": "cosine",
                        "compression": "LVQ4x4",
                        "graph_max_degree": 40,
                        "construction_window_size": 250,
                        "search_window_size": 20,
                    },
                },
            ],
        }
    )


@pytest.fixture
def svs_schema_leanvec(worker_id):
    """Create SVS-VAMANA schema with LeanVec compression and dimensionality reduction."""
    return IndexSchema.from_dict(
        {
            "index": {
                "name": f"svs_leanvec_{worker_id}",
                "prefix": f"svs_leanvec_{worker_id}",
            },
            "fields": [
                {"name": "id", "type": "tag"},
                {"name": "content", "type": "text"},
                {
                    "name": "embedding",
                    "type": "vector",
                    "attrs": {
                        "dims": 1536,
                        "algorithm": "svs-vamana",
                        "datatype": "float16",
                        "distance_metric": "cosine",
                        "compression": "LeanVec4x8",
                        "reduce": 768,
                        "graph_max_degree": 64,
                        "construction_window_size": 300,
                        "search_window_size": 30,
                    },
                },
            ],
        }
    )


@pytest.fixture
def svs_index_lvq(svs_schema_lvq, client):
    """Create SVS-VAMANA index with LVQ compression."""
    index = SearchIndex(schema=svs_schema_lvq, redis_client=client)
    index.create(overwrite=True)
    yield index
    index.delete(drop=True)


@pytest.fixture
def svs_index_leanvec(svs_schema_leanvec, client):
    """Create SVS-VAMANA index with LeanVec compression."""
    index = SearchIndex(schema=svs_schema_leanvec, redis_client=client)
    index.create(overwrite=True)
    yield index
    index.delete(drop=True)


def generate_test_vectors(dims, count=100, dtype="float32"):
    """Generate random test vectors."""
    vectors = []
    for i in range(count):
        vector = np.random.random(dims).astype(dtype)
        vectors.append(
            {
                "id": f"doc_{i}",
                "content": f"This is test document {i}",
                "embedding": array_to_buffer(vector, dtype=dtype),
            }
        )
    return vectors


class TestSVSCapabilityDetection:
    """Test SVS-VAMANA capability detection."""

    def test_check_svs_capabilities(self, client):
        """Test that SVS-VAMANA is supported on the test Redis instance."""
        # These tests require Redis 8.2+ with RediSearch 2.8.10+
        assert supports_svs(client) is True, (
            "SVS-VAMANA not supported. "
            "Requires Redis >= 8.2.0 with RediSearch >= 2.8.10"
        )


class TestSVSIndexCreation:
    """Test creating SVS-VAMANA indices with various configurations."""

    def test_create_svs_index_lvq(self, svs_index_lvq):
        """Test creating SVS-VAMANA index with LVQ compression."""
        assert svs_index_lvq.exists()

        # Verify index info
        info = svs_index_lvq.info()
        assert info["num_docs"] == 0

    def test_create_svs_index_leanvec(self, svs_index_leanvec):
        """Test creating SVS-VAMANA index with LeanVec compression."""
        assert svs_index_leanvec.exists()

        # Verify index info
        info = svs_index_leanvec.info()
        assert info["num_docs"] == 0

    def test_create_svs_with_compression_advisor(self, client, worker_id):
        """Test creating SVS-VAMANA index using CompressionAdvisor."""
        dims = 768
        config = CompressionAdvisor.recommend(dims=dims, priority="balanced")

        schema = IndexSchema.from_dict(
            {
                "index": {
                    "name": f"svs_advisor_{worker_id}",
                    "prefix": f"svs_advisor_{worker_id}",
                },
                "fields": [
                    {"name": "id", "type": "tag"},
                    {
                        "name": "embedding",
                        "type": "vector",
                        "attrs": {
                            "dims": dims,
                            **config,
                            "distance_metric": "cosine",
                        },
                    },
                ],
            }
        )

        index = SearchIndex(schema=schema, redis_client=client)
        index.create(overwrite=True)

        try:
            assert index.exists()
            info = index.info()
            assert info["num_docs"] == 0
        finally:
            index.delete(drop=True)


class TestSVSDataIngestion:
    """Test loading data into SVS-VAMANA indices."""

    def test_load_data_lvq(self, svs_index_lvq):
        """Test loading data into SVS-VAMANA index with LVQ compression."""
        vectors = generate_test_vectors(dims=384, count=50, dtype="float32")
        svs_index_lvq.load(vectors)

        # Verify data was loaded
        info = svs_index_lvq.info()
        assert info["num_docs"] == 50

    def test_load_data_leanvec(self, svs_index_leanvec):
        """Test loading data into SVS-VAMANA index with LeanVec compression."""
        vectors = generate_test_vectors(dims=1536, count=50, dtype="float32")
        svs_index_leanvec.load(vectors)

        # Verify data was loaded
        info = svs_index_leanvec.info()
        assert info["num_docs"] == 50

    def test_load_large_batch(self, svs_index_lvq):
        """Test loading larger batch of data."""
        vectors = generate_test_vectors(dims=384, count=200, dtype="float32")
        svs_index_lvq.load(vectors)

        # Verify data was loaded
        info = svs_index_lvq.info()
        assert info["num_docs"] == 200


class TestSVSQuerying:
    """Test querying SVS-VAMANA indices."""

    def test_vector_query_lvq(self, svs_index_lvq):
        """Test vector similarity search on SVS-VAMANA index with LVQ."""
        # Load test data
        vectors = generate_test_vectors(dims=384, count=100, dtype="float32")
        svs_index_lvq.load(vectors)

        # Create query vector
        query_vector = np.random.random(384).astype(np.float32)

        # Execute query
        query = VectorQuery(
            vector=query_vector,
            vector_field_name="embedding",
            return_fields=["id", "content"],
            num_results=10,
        )

        results = svs_index_lvq.query(query)

        # Verify results
        assert len(results) <= 10
        assert all("id" in result for result in results)
        assert all("content" in result for result in results)

    def test_vector_query_leanvec(self, svs_index_leanvec):
        """Test vector similarity search on SVS-VAMANA index with LeanVec."""
        # Load test data
        vectors = generate_test_vectors(dims=1536, count=100, dtype="float32")
        svs_index_leanvec.load(vectors)

        # Create query vector
        query_vector = np.random.random(1536).astype(np.float32)

        # Execute query
        query = VectorQuery(
            vector=query_vector,
            vector_field_name="embedding",
            return_fields=["id", "content"],
            num_results=5,
        )

        results = svs_index_leanvec.query(query)

        # Verify results
        assert len(results) <= 5
        assert all("id" in result for result in results)

    def test_query_with_filters(self, svs_index_lvq):
        """Test vector query with filters on SVS-VAMANA index."""
        # Load test data with specific IDs
        vectors = []
        for i in range(50):
            vector = np.random.random(384).astype(np.float32)
            vectors.append(
                {
                    "id": f"category_a_{i}" if i < 25 else f"category_b_{i}",
                    "content": f"Document {i}",
                    "embedding": array_to_buffer(vector, dtype="float32"),
                }
            )
        svs_index_lvq.load(vectors)

        # Query with filter
        query_vector = np.random.random(384).astype(np.float32)
        query = VectorQuery(
            vector=query_vector,
            vector_field_name="embedding",
            return_fields=["id", "content"],
            num_results=10,
            filter_expression="@id:{category_a*}",
        )

        results = svs_index_lvq.query(query)

        # Verify all results match filter
        assert len(results) <= 10
        assert all(result["id"].startswith("category_a") for result in results)


class TestSVSFromExisting:
    """Test loading existing SVS-VAMANA indices."""

    def test_from_existing_lvq(self, svs_index_lvq, client):
        """Test loading existing SVS-VAMANA index with LVQ compression."""
        # Load some data
        vectors = generate_test_vectors(dims=384, count=20, dtype="float32")
        svs_index_lvq.load(vectors)

        # Load the index from existing
        loaded_index = SearchIndex.from_existing(
            svs_index_lvq.name, redis_client=client
        )

        # Verify the loaded index
        assert loaded_index.exists()
        assert loaded_index.name == svs_index_lvq.name

        # Verify schema was loaded correctly
        embedding_field = loaded_index.schema.fields["embedding"]
        assert embedding_field.attrs.algorithm.value == "SVS-VAMANA"
        assert embedding_field.attrs.compression.value == "LVQ4x4"
        assert embedding_field.attrs.dims == 384

        # Verify data is accessible
        info = loaded_index.info()
        assert info["num_docs"] == 20

    def test_from_existing_leanvec(self, svs_index_leanvec, client):
        """Test loading existing SVS-VAMANA index with LeanVec compression."""
        # Load some data
        vectors = generate_test_vectors(dims=1536, count=20, dtype="float32")
        svs_index_leanvec.load(vectors)

        # Load the index from existing
        loaded_index = SearchIndex.from_existing(
            svs_index_leanvec.name, redis_client=client
        )

        # Verify the loaded index
        assert loaded_index.exists()
        assert loaded_index.name == svs_index_leanvec.name

        # Verify schema was loaded correctly
        embedding_field = loaded_index.schema.fields["embedding"]
        assert embedding_field.attrs.algorithm.value == "SVS-VAMANA"
        assert embedding_field.attrs.compression.value == "LeanVec4x8"
        assert embedding_field.attrs.dims == 1536
        assert embedding_field.attrs.reduce == 768

        # Verify data is accessible
        info = loaded_index.info()
        assert info["num_docs"] == 20


class TestSVSCompressionTypes:
    """Test different compression types for SVS-VAMANA."""

    @pytest.mark.parametrize(
        "compression,dims,dtype",
        [
            ("LVQ4", 384, "float32"),
            ("LVQ4x4", 384, "float32"),
            ("LVQ4x8", 384, "float32"),
            ("LVQ8", 384, "float32"),
        ],
    )
    def test_lvq_compression_types(self, client, worker_id, compression, dims, dtype):
        """Test various LVQ compression types."""
        schema = IndexSchema.from_dict(
            {
                "index": {
                    "name": f"svs_{compression.lower()}_{worker_id}",
                    "prefix": f"svs_{compression.lower()}_{worker_id}",
                },
                "fields": [
                    {"name": "id", "type": "tag"},
                    {
                        "name": "embedding",
                        "type": "vector",
                        "attrs": {
                            "dims": dims,
                            "algorithm": "svs-vamana",
                            "datatype": dtype,
                            "distance_metric": "cosine",
                            "compression": compression,
                        },
                    },
                ],
            }
        )

        index = SearchIndex(schema=schema, redis_client=client)
        index.create(overwrite=True)

        try:
            # Load data
            vectors = generate_test_vectors(dims=dims, count=50, dtype=dtype)
            index.load(vectors)

            # Verify
            assert index.exists()
            info = index.info()
            assert info["num_docs"] == 50

            # Query
            query_vector = np.random.random(dims).astype(dtype)
            query = VectorQuery(
                vector=query_vector,
                vector_field_name="embedding",
                return_fields=["id"],
                num_results=5,
            )
            results = index.query(query)
            assert len(results) <= 5
        finally:
            index.delete(drop=True)

    @pytest.mark.parametrize(
        "compression,dims,reduce,dtype",
        [
            ("LeanVec4x8", 1024, 512, "float16"),
            ("LeanVec4x8", 1536, 768, "float16"),
            ("LeanVec8x8", 1536, 768, "float16"),
        ],
    )
    def test_leanvec_compression_types(
        self, client, worker_id, compression, dims, reduce, dtype
    ):
        """Test various LeanVec compression types with dimensionality reduction."""
        schema = IndexSchema.from_dict(
            {
                "index": {
                    "name": f"svs_{compression.lower()}_{reduce}_{worker_id}",
                    "prefix": f"svs_{compression.lower()}_{reduce}_{worker_id}",
                },
                "fields": [
                    {"name": "id", "type": "tag"},
                    {
                        "name": "embedding",
                        "type": "vector",
                        "attrs": {
                            "dims": dims,
                            "algorithm": "svs-vamana",
                            "datatype": dtype,
                            "distance_metric": "cosine",
                            "compression": compression,
                            "reduce": reduce,
                        },
                    },
                ],
            }
        )

        index = SearchIndex(schema=schema, redis_client=client)
        index.create(overwrite=True)

        try:
            # Load data
            vectors = generate_test_vectors(dims=dims, count=50, dtype="float32")
            index.load(vectors)

            # Verify
            assert index.exists()
            info = index.info()
            assert info["num_docs"] == 50

            # Query
            query_vector = np.random.random(dims).astype(np.float32)
            query = VectorQuery(
                vector=query_vector,
                vector_field_name="embedding",
                return_fields=["id"],
                num_results=5,
            )
            results = index.query(query)
            assert len(results) <= 5
        finally:
            index.delete(drop=True)



"""Unit tests for UNF and NOINDEX field attributes."""

import pytest

from redisvl.schema.fields import (
    FieldFactory,
    GeoField,
    NumericField,
    TagField,
    TextField,
)


class TestTextFieldAttributes:
    """Test TextField support for no_index and unf attributes."""

    def test_no_index_attribute_with_sortable(self):
        """Test TextField with no_index=True and sortable=True."""
        field = TextField(name="title", attrs={"no_index": True, "sortable": True})

        assert field.attrs.no_index is True
        assert field.attrs.sortable is True

        redis_field = field.as_redis_field()
        args = redis_field.redis_args()
        assert "NOINDEX" in args
        assert "SORTABLE" in args

    def test_no_index_default_value(self):
        """Test TextField no_index defaults to False."""
        field = TextField(name="content")
        assert field.attrs.no_index is False

        redis_field = field.as_redis_field()
        args = redis_field.redis_args()
        assert "NOINDEX" not in args

    def test_unf_attribute_with_sortable(self):
        """Test TextField with unf=True and sortable=True."""
        field = TextField(name="title", attrs={"unf": True, "sortable": True})

        assert field.attrs.unf is True
        assert field.attrs.sortable is True

        redis_field = field.as_redis_field()
        args = redis_field.redis_args()
        assert "UNF" in args
        assert "SORTABLE" in args

    def test_unf_without_sortable(self):
        """Test that UNF is not added when field is not sortable."""
        field = TextField(name="description", attrs={"unf": True, "sortable": False})

        assert field.attrs.unf is True
        assert field.attrs.sortable is False

        redis_field = field.as_redis_field()
        args = redis_field.redis_args()
        assert "UNF" not in args
        assert "SORTABLE" not in args

    def test_unf_default_value(self):
        """Test TextField unf defaults to False."""
        field = TextField(name="content")
        assert field.attrs.unf is False


class TestNumericFieldAttributes:
    """Test NumericField support for no_index and unf attributes."""

    def test_no_index_attribute_with_sortable(self):
        """Test NumericField with no_index=True and sortable=True."""
        field = NumericField(name="price", attrs={"no_index": True, "sortable": True})

        assert field.attrs.no_index is True
        assert field.attrs.sortable is True

        redis_field = field.as_redis_field()
        args = redis_field.redis_args()
        assert "NOINDEX" in args
        assert "SORTABLE" in args

    def test_no_index_default_value(self):
        """Test NumericField no_index defaults to False."""
        field = NumericField(name="quantity")
        assert field.attrs.no_index is False

        redis_field = field.as_redis_field()
        args = redis_field.redis_args()
        assert "NOINDEX" not in args

    def test_unf_attribute_with_sortable(self):
        """Test NumericField with unf=True and sortable=True."""
        field = NumericField(name="score", attrs={"unf": True, "sortable": True})

        assert field.attrs.unf is True
        assert field.attrs.sortable is True

        redis_field = field.as_redis_field()
        args = redis_field.redis_args()
        assert "UNF" in args
        assert "SORTABLE" in args

    def test_unf_without_sortable(self):
        """Test that UNF is not added when field is not sortable."""
        field = NumericField(name="count", attrs={"unf": True, "sortable": False})

        assert field.attrs.unf is True
        assert field.attrs.sortable is False

        redis_field = field.as_redis_field()
        args = redis_field.redis_args()
        assert "UNF" not in args
        assert "SORTABLE" not in args

    def test_unf_default_value(self):
        """Test NumericField unf defaults to False."""
        field = NumericField(name="rating")
        assert field.attrs.unf is False


class TestTagFieldNoIndex:
    """Test TagField support for no_index attribute."""

    def test_no_index_attribute_with_sortable(self):
        """Test TagField with no_index=True and sortable=True."""
        field = TagField(name="tags", attrs={"no_index": True, "sortable": True})

        assert field.attrs.no_index is True
        assert field.attrs.sortable is True

        redis_field = field.as_redis_field()
        args = redis_field.redis_args()
        assert "NOINDEX" in args
        assert "SORTABLE" in args

    def test_no_index_default_value(self):
        """Test TagField no_index defaults to False."""
        field = TagField(name="categories")
        assert field.attrs.no_index is False

        redis_field = field.as_redis_field()
        args = redis_field.redis_args()
        assert "NOINDEX" not in args


class TestGeoFieldNoIndex:
    """Test GeoField support for no_index attribute."""

    def test_no_index_attribute_with_sortable(self):
        """Test GeoField with no_index=True and sortable=True."""
        field = GeoField(name="location", attrs={"no_index": True, "sortable": True})

        assert field.attrs.no_index is True
        assert field.attrs.sortable is True

        redis_field = field.as_redis_field()
        args = redis_field.redis_args()
        assert "NOINDEX" in args
        assert "SORTABLE" in args

    def test_no_index_default_value(self):
        """Test GeoField no_index defaults to False."""
        field = GeoField(name="coordinates")
        assert field.attrs.no_index is False

        redis_field = field.as_redis_field()
        args = redis_field.redis_args()
        assert "NOINDEX" not in args


class TestFieldFactoryWithNewAttributes:
    """Test FieldFactory creating fields with new attributes."""

    def test_create_text_field_with_unf_and_noindex(self):
        """Test creating TextField with unf and no_index via FieldFactory."""
        field = FieldFactory.create_field(
            type="text",
            name="title",
            attrs={"unf": True, "no_index": True, "sortable": True},
        )

        assert isinstance(field, TextField)
        assert field.attrs.unf is True
        assert field.attrs.no_index is True
        assert field.attrs.sortable is True

    def test_create_numeric_field_with_unf_and_noindex(self):
        """Test creating NumericField with unf and no_index via FieldFactory."""
        field = FieldFactory.create_field(
            type="numeric",
            name="score",
            attrs={"unf": True, "no_index": True, "sortable": True},
        )

        assert isinstance(field, NumericField)
        assert field.attrs.unf is True
        assert field.attrs.no_index is True
        assert field.attrs.sortable is True

    def test_create_tag_field_with_noindex(self):
        """Test creating TagField with no_index via FieldFactory."""
        field = FieldFactory.create_field(
            type="tag", name="tags", attrs={"no_index": True, "sortable": True}
        )

        assert isinstance(field, TagField)
        assert field.attrs.no_index is True
        assert field.attrs.sortable is True

    def test_create_geo_field_with_noindex(self):
        """Test creating GeoField with no_index via FieldFactory."""
        field = FieldFactory.create_field(
            type="geo", name="location", attrs={"no_index": True, "sortable": True}
        )

        assert isinstance(field, GeoField)
        assert field.attrs.no_index is True
        assert field.attrs.sortable is True


class TestBackwardCompatibility:
    """Test that new attributes don't break backward compatibility."""

    def test_text_field_without_new_attributes(self):
        """Test TextField works without specifying new attributes."""
        field = TextField(name="content", attrs={"weight": 2.0})

        assert field.attrs.weight == 2.0
        assert field.attrs.unf is False
        assert field.attrs.no_index is False

    def test_numeric_field_without_new_attributes(self):
        """Test NumericField works without specifying new attributes."""
        field = NumericField(name="price", attrs={"sortable": True})

        assert field.attrs.sortable is True
        assert field.attrs.unf is False
        assert field.attrs.no_index is False

    def test_tag_field_without_new_attributes(self):
        """Test TagField works without specifying new attributes."""
        field = TagField(name="tags", attrs={"separator": "|"})

        assert field.attrs.separator == "|"
        assert field.attrs.no_index is False

    def test_geo_field_without_new_attributes(self):
        """Test GeoField works without specifying new attributes."""
        field = GeoField(name="location", attrs={"sortable": True})

        assert field.attrs.sortable is True
        assert field.attrs.no_index is False



from typing import Dict, List, Optional, Set, Tuple, Union

from redisvl.query.filter import FilterExpression
from redisvl.utils.token_escaper import TokenEscaper
from redisvl.utils.utils import lazy_import

nltk = lazy_import("nltk")
nltk_stopwords = lazy_import("nltk.corpus.stopwords")


def _parse_text_weights(weights: Optional[Dict[str, float]]) -> Dict[str, float]:
    parsed_weights: Dict[str, float] = {}
    if not weights:
        return parsed_weights
    for word, weight in weights.items():
        word = word.strip().lower()
        if not word or " " in word:
            raise ValueError(
                f"Only individual words may be weighted. Got {{ {word}:{weight} }}"
            )
        if not (isinstance(weight, float) or isinstance(weight, int)) or weight < 0.0:
            raise ValueError(
                f"Weights must be positive number. Got {{ {word}:{weight} }}"
            )
        parsed_weights[word] = weight
    return parsed_weights


class FullTextQueryHelper:
    """Convert raw user queries into Redis full-text queries - tokenizes, escapes, and filters stopwords from the query."""

    def __init__(
        self,
        stopwords: Optional[Union[str, Set[str]]] = "english",
        text_weights: Optional[Dict[str, float]] = None,
    ):
        self._stopwords = self._get_stopwords(stopwords)
        self._text_weights = _parse_text_weights(text_weights)

    @property
    def stopwords(self) -> Set[str]:
        """Return the stopwords used in the query.
        Returns:
            Set[str]: The stopwords used in the query.
        """
        return self._stopwords.copy() if self._stopwords else set()

    @property
    def text_weights(self) -> Dict[str, float]:
        """Get the text weights.

        Returns:
            Dictionary of word:weight mappings.
        """
        return self._text_weights

    def build_query_string(
        self,
        text: str,
        text_field_name: str,
        filter_expression: Optional[Union[str, FilterExpression]] = None,
    ) -> str:
        """Build the full-text query string for text search with optional filtering."""
        if isinstance(filter_expression, FilterExpression):
            filter_expression = str(filter_expression)

        query = f"(~@{text_field_name}:({self._tokenize_and_escape_query(text)})"

        if filter_expression and filter_expression != "*":
            query += f" AND {filter_expression}"

        return query + ")"

    def _get_stopwords(
        self, stopwords: Optional[Union[str, Set[str]]] = "english"
    ) -> Set[str]:
        """Get the stopwords to use in the query.

        Args:
            stopwords (Optional[Union[str, Set[str]]]): The stopwords to use. If a string
                such as "english" "german" is provided then a default set of stopwords for that
                language will be used. if a list, set, or tuple of strings is provided then those
                will be used as stopwords. Defaults to "english". if set to "None" then no stopwords
                will be removed.

        Returns:
            The set of stopwords to use.

        Raises:
            TypeError: If the stopwords are not a set, list, or tuple of strings.
        """
        if not stopwords:
            return set()
        elif isinstance(stopwords, str):
            try:
                # Try loading first; only download if not already present.
                # This avoids race conditions when parallel workers (e.g.
                # pytest-xdist) call nltk.download() concurrently.
                try:
                    return set(nltk_stopwords.words(stopwords))
                except LookupError:
                    nltk.download("stopwords", quiet=True)
                    return set(nltk_stopwords.words(stopwords))
            except ImportError:
                raise ValueError(
                    f"Loading stopwords for {stopwords} failed: nltk is not installed."
                )
            except Exception as e:
                raise ValueError(f"Error trying to load {stopwords} from nltk. {e}")
        elif isinstance(stopwords, (Set, List, Tuple)) and all(  # type: ignore
            isinstance(word, str) for word in stopwords
        ):
            return set(stopwords)
        else:
            raise TypeError("stopwords must be a set, list, or tuple of strings")

    def set_text_weights(self, weights: Dict[str, float]):
        """Set or update the text weights for the query.

        Args:
            weights: Dictionary of word:weight mappings
        """
        self._text_weights = _parse_text_weights(weights)

    def _tokenize_and_escape_query(self, user_query: str) -> str:
        """Convert a raw user query to a redis full text query joined by ORs

        Args:
            user_query (str): The user query to tokenize and escape.

        Returns:
            str: The tokenized and escaped query string.

        Raises:
            ValueError: If the text string becomes empty after stopwords are removed.
        """
        escaper = TokenEscaper()

        tokens = [
            escaper.escape(
                token.strip().strip(",").replace("“", "").replace("”", "").lower()
            )
            for token in user_query.split()
        ]

        token_list = [
            token for token in tokens if token and token not in self._stopwords
        ]
        for i, token in enumerate(token_list):
            if token in self._text_weights:
                token_list[i] = f"{token}=>{{$weight:{self._text_weights[token]}}}"

        if not token_list:
            raise ValueError("text string cannot be empty after removing stopwords")
        return " | ".join(token_list)



import warnings
from random import choice
from unittest import mock

import pytest
from redis import Redis as SyncRedis
from redis.asyncio import Redis as AsyncRedis

from redisvl.exceptions import QueryValidationError, RedisSearchError, RedisVLError
from redisvl.index import AsyncSearchIndex
from redisvl.query import VectorQuery
from redisvl.query.query import FilterQuery
from redisvl.redis.utils import convert_bytes
from redisvl.schema import IndexSchema, StorageType
from redisvl.schema.fields import VectorIndexAlgorithm

fields = [{"name": "test", "type": "tag"}]


@pytest.fixture
def index_schema(worker_id):
    return IndexSchema.from_dict(
        {"index": {"name": f"my_index_{worker_id}"}, "fields": fields}
    )


@pytest.fixture
def async_index(index_schema, async_client):
    return AsyncSearchIndex(schema=index_schema, redis_client=async_client)


@pytest.fixture
def async_index_from_dict(worker_id):

    return AsyncSearchIndex.from_dict(
        {
            "index": {"name": f"my_index_{worker_id}", "prefix": f"rvl_{worker_id}"},
            "fields": fields,
        }
    )


@pytest.fixture
def async_index_from_yaml(worker_id):

    index = AsyncSearchIndex.from_yaml("schemas/test_json_schema.yaml")
    # Update the index name and prefix to include worker_id
    index.schema.index.name = f"{index.schema.index.name}_{worker_id}"
    index.schema.index.prefix = f"{index.schema.index.prefix}_{worker_id}"
    return index


def test_search_index_properties(index_schema, async_index):
    assert async_index.schema == index_schema
    # custom settings
    assert async_index.name == index_schema.index.name
    assert async_index.name.startswith("my_index_")
    assert async_index.client
    # default settings
    assert async_index.prefix == index_schema.index.prefix == "rvl"
    assert async_index.key_separator == index_schema.index.key_separator == ":"
    assert (
        async_index.storage_type == index_schema.index.storage_type == StorageType.HASH
    )
    assert async_index.key("foo").startswith(async_index.prefix)


def test_search_index_from_yaml(async_index_from_yaml):
    assert async_index_from_yaml.name.startswith("json-test")
    assert async_index_from_yaml.client is None
    assert async_index_from_yaml.prefix.startswith("json_")
    assert async_index_from_yaml.key_separator == ":"
    assert async_index_from_yaml.storage_type == StorageType.JSON
    assert async_index_from_yaml.key("foo").startswith(async_index_from_yaml.prefix)


def test_search_index_from_dict(async_index_from_dict):
    assert async_index_from_dict.name.startswith("my_index")
    assert async_index_from_dict.client is None
    assert async_index_from_dict.prefix.startswith("rvl_")
    assert async_index_from_dict.key_separator == ":"
    assert async_index_from_dict.storage_type == StorageType.HASH
    assert len(async_index_from_dict.schema.fields) == len(fields)
    assert async_index_from_dict.key("foo").startswith(async_index_from_dict.prefix)


@pytest.mark.asyncio
async def test_search_index_from_existing(async_client, async_index):
    await async_index.create(overwrite=True)

    try:
        async_index2 = await AsyncSearchIndex.from_existing(
            async_index.name, redis_client=async_client
        )
    except Exception as e:
        pytest.skip(str(e))

    assert async_index2.schema == async_index.schema


@pytest.mark.asyncio
async def test_search_index_from_existing_url(async_index, redis_url):
    await async_index.create(overwrite=True)

    try:
        async_index2 = await AsyncSearchIndex.from_existing(
            async_index.name, redis_url=redis_url
        )
    except Exception as e:
        pytest.skip(str(e))

    assert async_index2.schema == async_index.schema


@pytest.mark.asyncio
async def test_search_index_from_existing_complex(async_client):
    schema = {
        "index": {
            "name": "test",
            "prefix": "test",
            "storage_type": "json",
        },
        "fields": [
            {"name": "user", "type": "tag", "path": "$.user"},
            {"name": "credit_score", "type": "tag", "path": "$.metadata.credit_score"},
            {"name": "job", "type": "text", "path": "$.metadata.job"},
            {
                "name": "age",
                "type": "numeric",
                "path": "$.metadata.age",
                "attrs": {"sortable": False},
            },
            {
                "name": "user_embedding",
                "type": "vector",
                "attrs": {
                    "dims": 3,
                    "distance_metric": "cosine",
                    "algorithm": "flat",
                    "datatype": "float32",
                },
            },
        ],
    }
    async_index = AsyncSearchIndex.from_dict(schema, redis_client=async_client)
    await async_index.create(overwrite=True)

    try:
        async_index2 = await AsyncSearchIndex.from_existing(
            async_index.name, redis_client=async_client
        )
    except Exception as e:
        pytest.skip(str(e))

    # Verify index metadata matches
    assert async_index2.schema.index.name == async_index.schema.index.name
    assert async_index2.schema.index.prefix == async_index.schema.index.prefix
    assert (
        async_index2.schema.index.storage_type == async_index.schema.index.storage_type
    )

    # Verify non-vector fields are present
    for field_name in ["user", "credit_score", "job", "age"]:
        assert field_name in async_index2.schema.fields
        assert (
            async_index2.schema.fields[field_name].type
            == async_index.schema.fields[field_name].type
        )

    # Vector field may not be present on older Redis versions
    if "user_embedding" in async_index2.schema.fields:
        assert async_index2.schema.fields["user_embedding"].type == "vector"


def test_search_index_no_prefix(index_schema):
    # specify an explicitly empty prefix...
    index_schema.index.prefix = ""
    async_index = AsyncSearchIndex(schema=index_schema)
    assert async_index.prefix == ""
    assert async_index.key("foo") == "foo"


@pytest.mark.asyncio
async def test_search_index_redis_url(redis_url, index_schema):
    async with AsyncSearchIndex(
        schema=index_schema, redis_url=redis_url
    ) as async_index:
        # Client is None until a command is run
        assert async_index.client is None

        # Lazily create the client by running a command
        await async_index.create(overwrite=True, drop=True)
        assert async_index.client

    # After exiting async with, if the index owned the client, it should be disconnected
    # and client attribute should be None again by __aexit__
    assert async_index.client is None


@pytest.mark.asyncio
async def test_search_index_client(async_client, index_schema):
    async_index = AsyncSearchIndex(schema=index_schema, redis_client=async_client)
    assert async_index.client == async_client


@pytest.mark.asyncio
async def test_search_index_set_client(client, redis_url, index_schema):
    # Use async with for the index that owns its initial client via redis_url
    async with AsyncSearchIndex(
        schema=index_schema, redis_url=redis_url
    ) as async_index:
        # Ignore deprecation warnings for set_client
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            await async_index.create(overwrite=True, drop=True)
            assert isinstance(async_index.client, AsyncRedis)

            # Tests deprecated sync -> async conversion behavior
            assert isinstance(client, SyncRedis)

            await async_index.set_client(client)
            assert isinstance(async_index.client, AsyncRedis)

            if async_index.client:
                await async_index.disconnect()
            assert async_index.client is None


@pytest.mark.asyncio
async def test_search_index_create(async_index):
    await async_index.create(overwrite=True, drop=True)
    assert await async_index.exists()
    assert async_index.name in convert_bytes(
        await async_index.client.execute_command("FT._LIST")
    )


@pytest.mark.asyncio
async def test_search_index_delete(async_index):
    await async_index.create(overwrite=True, drop=True)
    await async_index.delete(drop=True)
    assert not await async_index.exists()
    assert async_index.name not in convert_bytes(
        await async_index.client.execute_command("FT._LIST")
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("num_docs", [0, 1, 5, 10, 2042])
async def test_search_index_clear(async_index, num_docs):
    await async_index.create(overwrite=True, drop=True)
    tags = ["foo", "bar", "baz"]
    data = [{"id": str(i), "test": choice(tags)} for i in range(num_docs)]
    await async_index.load(data, id_field="id")
    info = await async_index.info()
    assert info["num_docs"] == num_docs

    count = await async_index.clear()
    assert count == num_docs
    assert await async_index.exists()
    info = await async_index.info()
    assert info["num_docs"] == 0


@pytest.mark.asyncio
async def test_search_index_drop_key(async_index):
    await async_index.create(overwrite=True, drop=True)
    data = [{"id": "1", "test": "foo"}, {"id": "2", "test": "bar"}]
    keys = await async_index.load(data, id_field="id")

    dropped = await async_index.drop_keys(keys[0])
    assert dropped == 1
    assert not await async_index.fetch(keys[0])
    assert await async_index.fetch(keys[1]) is not None


@pytest.mark.asyncio
async def test_search_index_drop_keys(async_index):
    await async_index.create(overwrite=True, drop=True)
    data = [
        {"id": "1", "test": "foo"},
        {"id": "2", "test": "bar"},
        {"id": "3", "test": "baz"},
    ]
    keys = await async_index.load(data, id_field="id")

    dropped = await async_index.drop_keys(keys[0:2])
    assert dropped == 2
    assert not await async_index.fetch(keys[0])
    assert not await async_index.fetch(keys[1])
    assert await async_index.fetch(keys[2]) is not None

    assert await async_index.exists()


@pytest.mark.asyncio
async def test_search_index_drop_documents(async_index):
    await async_index.create(overwrite=True, drop=True)
    data = [
        {"id": "1", "test": "foo"},
        {"id": "2", "test": "bar"},
        {"id": "3", "test": "baz"},
    ]
    await async_index.load(data, id_field="id")

    # Test dropping a single document by ID
    dropped = await async_index.drop_documents("1")
    assert dropped == 1
    assert not await async_index.fetch("1")
    assert await async_index.fetch("2") is not None
    assert await async_index.fetch("3") is not None

    # Test dropping multiple documents by ID
    dropped = await async_index.drop_documents(["2", "3"])
    assert dropped == 2
    assert not await async_index.fetch("2")
    assert not await async_index.fetch("3")

    # Test dropping with an empty list
    dropped = await async_index.drop_documents([])
    assert dropped == 0

    # Ensure the index still exists
    assert await async_index.exists()


@pytest.mark.asyncio
async def test_search_index_load_and_fetch(async_index):
    await async_index.create(overwrite=True, drop=True)
    data = [{"id": "1", "test": "foo"}]
    await async_index.load(data, id_field="id")

    res = await async_index.fetch("1")
    assert (
        res["test"]
        == convert_bytes(await async_index.client.hget("rvl:1", "test"))
        == "foo"
    )

    await async_index.delete(drop=True)
    assert not await async_index.exists()
    assert not await async_index.fetch("1")


@pytest.mark.asyncio
async def test_search_index_load_preprocess(async_index):
    await async_index.create(overwrite=True, drop=True)
    data = [{"id": "1", "test": "foo"}]

    def preprocess(record):
        record["test"] = "bar"
        return record

    await async_index.load(data, id_field="id", preprocess=preprocess)
    res = await async_index.fetch("1")
    assert (
        res["test"]
        == convert_bytes(await async_index.client.hget("rvl:1", "test"))
        == "bar"
    )

    def bad_preprocess(record):
        return 1

    with pytest.raises(RedisVLError):
        await async_index.load(data, id_field="id", preprocess=bad_preprocess)


@pytest.mark.asyncio
async def test_search_index_load_empty(async_index):
    await async_index.create(overwrite=True, drop=True)
    await async_index.load([])


@pytest.mark.asyncio
async def test_no_id_field(async_index):
    await async_index.create(overwrite=True, drop=True)
    bad_data = [{"wrong_key": "1", "value": "test"}]

    # catch missing / invalid id_field
    with pytest.raises(RedisVLError):
        await async_index.load(bad_data, id_field="key")


@pytest.mark.asyncio
async def test_check_index_exists_before_delete(async_index):
    await async_index.create(overwrite=True, drop=True)
    await async_index.delete(drop=True)
    with pytest.raises(RedisSearchError):
        await async_index.delete()


@pytest.mark.asyncio
async def test_check_index_exists_before_search(async_index):
    await async_index.create(overwrite=True, drop=True)
    await async_index.delete(drop=True)

    query = VectorQuery(
        [0.1, 0.1, 0.5],
        "user_embedding",
        return_fields=["user", "credit_score", "age", "job", "location"],
        num_results=7,
    )
    with pytest.raises(RedisSearchError):
        await async_index.search(query.query, query_params=query.params)


@pytest.mark.asyncio
async def test_check_index_exists_before_info(async_index):
    await async_index.create(overwrite=True, drop=True)
    await async_index.delete(drop=True)

    with pytest.raises(RedisSearchError):
        await async_index.info()


@pytest.mark.asyncio
async def test_search_index_that_does_not_own_client_context_manager(async_index):
    async with async_index:
        await async_index.create(overwrite=True, drop=True)
        assert async_index._redis_client
        client = async_index._redis_client
    assert async_index._redis_client == client


@pytest.mark.asyncio
async def test_search_index_that_does_not_own_client_context_manager_with_exception(
    async_index,
):
    try:
        async with async_index:
            await async_index.create(overwrite=True, drop=True)
            client = async_index._redis_client
            raise ValueError("test")
    except ValueError:
        pass
    assert async_index._redis_client == client


@pytest.mark.asyncio
async def test_search_index_that_does_not_own_client_disconnect(async_index):
    await async_index.create(overwrite=True, drop=True)
    client = async_index._redis_client
    await async_index.disconnect()
    assert async_index._redis_client == client


@pytest.mark.asyncio
async def test_search_index_that_does_not_own_client_disconnect_sync(async_index):
    await async_index.create(overwrite=True, drop=True)
    client = async_index._redis_client
    async_index.disconnect_sync()
    assert async_index._redis_client == client


@pytest.mark.asyncio
async def test_search_index_that_owns_client_context_manager(index_schema, redis_url):
    async_index = AsyncSearchIndex(schema=index_schema, redis_url=redis_url)
    async with async_index:
        await async_index.create(overwrite=True, drop=True)
        assert async_index._redis_client
    assert async_index._redis_client is None


@pytest.mark.asyncio
async def test_search_index_that_owns_client_context_manager_with_exception(
    index_schema, redis_url
):
    async_index = AsyncSearchIndex(schema=index_schema, redis_url=redis_url)
    try:
        async with async_index:
            await async_index.create(overwrite=True, drop=True)
            raise ValueError("test")
    except ValueError:
        pass
    assert async_index._redis_client is None


@pytest.mark.asyncio
async def test_search_index_that_owns_client_disconnect(index_schema, redis_url):
    async_index = AsyncSearchIndex(schema=index_schema, redis_url=redis_url)
    await async_index.create(overwrite=True, drop=True)
    await async_index.disconnect()
    assert async_index._redis_client is None


@pytest.mark.asyncio
async def test_search_index_that_owns_client_disconnect_sync(index_schema, redis_url):
    async_index = AsyncSearchIndex(schema=index_schema, redis_url=redis_url)
    await async_index.create(overwrite=True, drop=True)
    await async_index.disconnect()
    assert async_index._redis_client is None


@pytest.mark.asyncio
async def test_async_search_index_no_proactive_module_validation(redis_url):
    """
    Updated test for issue #370: AsyncSearchIndex should not validate modules proactively.
    Operations should fail naturally if modules are missing.
    """
    client = AsyncRedis.from_url(redis_url)
    with mock.patch(
        "redisvl.index.index.RedisConnectionFactory.validate_async_redis"
    ) as mock_validate_async_redis:
        # Create index - validation should only set lib name, not check modules
        index = AsyncSearchIndex(
            schema=IndexSchema.from_dict(
                {"index": {"name": "my_index"}, "fields": fields}
            ),
            redis_client=client,
        )

        # Access client to trigger lazy init
        _ = await index._get_client()

        # validate_async_redis might be called to set lib name, but won't raise module errors
        # The actual operation (create) will succeed if modules are present
        await index.create(overwrite=True, drop=True)

        # Verify index was created successfully (modules are present in test env)
        assert await index.exists()


@pytest.mark.asyncio
async def test_batch_search(async_index):
    await async_index.create(overwrite=True, drop=True)
    data = [{"id": "1", "test": "foo"}, {"id": "2", "test": "bar"}]
    await async_index.load(data, id_field="id")

    results = await async_index.batch_search(["@test:{foo}", "@test:{bar}"])
    assert len(results) == 2
    assert results[0].total == 1
    assert results[0].docs[0]["id"] == "rvl:1"
    assert results[1].total == 1
    assert results[1].docs[0]["id"] == "rvl:2"


@pytest.mark.parametrize(
    "queries",
    [
        [
            [
                FilterQuery(filter_expression="@test:{foo}"),
                FilterQuery(filter_expression="@test:{bar}"),
            ],
            [
                FilterQuery(filter_expression="@test:{foo}"),
                FilterQuery(filter_expression="@test:{bar}"),
                FilterQuery(filter_expression="@test:{baz}"),
                FilterQuery(filter_expression="@test:{foo}"),
                FilterQuery(filter_expression="@test:{bar}"),
                FilterQuery(filter_expression="@test:{baz}"),
            ],
        ],
        [
            [
                "@test:{foo}",
                "@test:{bar}",
            ],
            [
                "@test:{foo}",
                "@test:{bar}",
                "@test:{baz}",
                "@test:{foo}",
                "@test:{bar}",
                "@test:{baz}",
            ],
        ],
    ],
)
@pytest.mark.asyncio
async def test_batch_search_with_multiple_batches(async_index, queries):
    await async_index.create(overwrite=True, drop=True)
    data = [{"id": "1", "test": "foo"}, {"id": "2", "test": "bar"}]
    await async_index.load(data, id_field="id")

    results = await async_index.batch_search(queries[0])
    assert len(results) == 2
    assert results[0].total == 1
    assert results[0].docs[0]["id"] == "rvl:1"
    assert results[1].total == 1
    assert results[1].docs[0]["id"] == "rvl:2"

    results = await async_index.batch_search(
        queries[1],
        batch_size=2,
    )
    assert len(results) == 6

    # First (and only) result for the first query
    assert results[0].total == 1
    assert results[0].docs[0]["id"] == "rvl:1"

    # Second (and only) result for the second query
    assert results[1].total == 1
    assert results[1].docs[0]["id"] == "rvl:2"

    # Third query should have zero results because there is no baz
    assert results[2].total == 0

    # Then the pattern repeats
    assert results[3].total == 1
    assert results[3].docs[0]["id"] == "rvl:1"
    assert results[4].total == 1
    assert results[4].docs[0]["id"] == "rvl:2"
    assert results[5].total == 0


@pytest.mark.asyncio
async def test_batch_query(async_index):
    await async_index.create(overwrite=True, drop=True)
    data = [{"id": "1", "test": "foo"}, {"id": "2", "test": "bar"}]
    await async_index.load(data, id_field="id")

    query = FilterQuery(filter_expression="@test:{foo}")
    results = await async_index.batch_query([query])

    assert len(results) == 1
    assert results[0][0]["id"] == "rvl:1"


@pytest.mark.asyncio
async def test_batch_query_with_multiple_batches(async_index):
    await async_index.create(overwrite=True, drop=True)
    data = [{"id": "1", "test": "foo"}, {"id": "2", "test": "bar"}]
    await async_index.load(data, id_field="id")

    queries = [
        FilterQuery(filter_expression="@test:{foo}"),
        FilterQuery(filter_expression="@test:{bar}"),
    ]
    results = await async_index.batch_query(queries, batch_size=1)
    assert len(results) == 2
    assert results[0][0]["id"] == "rvl:1"
    assert results[1][0]["id"] == "rvl:2"


@pytest.mark.asyncio
async def test_async_search_index_expire_keys(async_index):
    """Test that AsyncSearchIndex.expire_keys method properly sets expiration times on keys."""
    # Create the index and load some data
    await async_index.create(overwrite=True, drop=True)
    data = [{"id": "1", "test": "foo"}, {"id": "2", "test": "bar"}]
    keys = await async_index.load(data, id_field="id")

    # Set expiration on a single key
    await async_index.expire_keys(keys[0], 60)
    client = await async_index._get_client()
    ttl = await client.ttl(keys[0])
    assert ttl > 0  # TTL should be positive
    assert ttl <= 60  # TTL should be 60 or less

    # Test no expiration on the other key
    ttl = await client.ttl(keys[1])
    assert ttl == -1  # -1 means no expiration

    # Set expiration on multiple keys
    result = await async_index.expire_keys(keys, 30)
    assert isinstance(result, list)
    assert len(result) == 2
    assert all(r == 1 for r in result)  # All operations should return 1 (success)

    # Verify TTLs are set
    for key in keys:
        ttl = await client.ttl(key)
        assert ttl > 0
        assert ttl <= 30


@pytest.mark.asyncio
async def test_search_index_validates_query_with_flat_algorithm(
    async_flat_index, sample_data
):
    assert (
        async_flat_index.schema.fields["user_embedding"].attrs.algorithm
        == VectorIndexAlgorithm.FLAT
    )
    query = VectorQuery(
        [0.1, 0.1, 0.5],
        "user_embedding",
        return_fields=["user", "credit_score", "age", "job", "location"],
        num_results=7,
        ef_runtime=100,
    )
    with pytest.raises(QueryValidationError):
        await async_flat_index.query(query)


@pytest.mark.asyncio
async def test_search_index_validates_query_with_hnsw_algorithm(
    async_hnsw_index, sample_data
):
    assert (
        async_hnsw_index.schema.fields["user_embedding"].attrs.algorithm
        == VectorIndexAlgorithm.HNSW
    )
    query = VectorQuery(
        [0.1, 0.1, 0.5],
        "user_embedding",
        return_fields=["user", "credit_score", "age", "job", "location"],
        num_results=7,
        ef_runtime=100,
    )
    # Should not raise
    await async_hnsw_index.query(query)


@pytest.mark.asyncio
async def test_async_search_index_connect(index_schema, redis_url):
    """Test that AsyncSearchIndex.connect() works with redis_url parameter."""
    async_index = AsyncSearchIndex(schema=index_schema)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        await async_index.connect(redis_url=redis_url)
    assert async_index.client is not None
    await async_index.disconnect()


@pytest.mark.asyncio
@pytest.mark.parametrize("ttl", [None, 30])
async def test_search_index_load_with_ttl(async_index, ttl):
    """Test that TTL is correctly set on keys when using load() with ttl parameter."""
    await async_index.create(overwrite=True, drop=True)

    # Load test data with TTL parameter
    data = [{"id": "1", "test": "foo"}]
    keys = await async_index.load(data, id_field="id", ttl=ttl)

    # Check TTL on the loaded key
    client = await async_index._get_client()
    key_ttl = await client.ttl(keys[0])

    if ttl is None:
        # No TTL set, should return -1
        assert key_ttl == -1
    else:
        # TTL should be set and close to the expected value
        assert key_ttl > 0
        assert abs(key_ttl - ttl) <= 5
